#---------------------------------------------------------------------------------------------------

bl_info = {
    "name": "NOX Renderer",
    "author": "",
    "version": (1,6),
    "blender": (2, 5, 8),
    "api": 36212,
    "location": "Render > Engine > NOX Renderer",
    "description": "Export scene and run NOX Renderer",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Render"}


#---------------------------------------------------------------------------------------------------

import bpy
from render_nox.debug import OkOperator, MessageOperator
from render_nox.render.render_nox import NOXRenderEngine
from render_nox.material.material_props import NOXTextureSlotProps, NOXMatLayerProps, NOXMaterialProps
from render_nox.material.material_panels import TestMatPanel, NOXTexturePanel
from render_nox.material.material_ops import RunMatEditorOperator, add_nox_image_texture, change_tex_type_to_image, uncheck_tex_in_nox_slot, add_remove_material_layer, delete_tex_permanently
from render_nox.material.material_import import NOXMaterialImporter
from render_nox.material.material_export import NOXMaterialExporter
from render_nox.export.export_nox_scene import ExportNOXScene
from render_nox.camera.camera_panels import NOXCameraPanel
from render_nox.camera.camera_props import NOXCameraProps
from render_nox.debug import NOX_debug
from render_nox.render.render_panels import NOXRenderPanel
from render_nox.render.render_ops import RunNOXRendererOperator
from render_nox.scene.scene_props import NOXSceneProps, NOXBlendLayersProps
from render_nox.render.render_panels import NOXBlendLayersPanel
from render_nox.env.sunsky_panels import NOXSunskyPanel, NOXEnvmapPanel
from render_nox.env.sunsky_props import NOXSunskyProps, NOXEnvMapProps
from render_nox.env.sunsky_ops import eval_sun_pos
from render_nox.camera.camera_ops import change_camera_length, set_camera_active
from render_nox.tools.tools_ops import DrawSelectedBoundsOperator, DrawSelectedSolidOperator
from render_nox.tools.tools_panels import NOXToolsPanel



def register():
    logON = NOX_debug.log_start_stop_nox_addon

    NOX_debug.addConsoleLog(logON, "----------------------------------------------------------")
    NOX_debug.addConsoleLog(logON, "Starting NOX Renderer add-on")
    
    NOX_debug.addConsoleLog(logON, "registering message box operator")
    bpy.utils.register_class(OkOperator)
    bpy.utils.register_class(MessageOperator)
    
    NOX_debug.addConsoleLog(logON, "registering camera properties")
    bpy.utils.register_class(NOXCameraProps)
    
    NOX_debug.addConsoleLog(logON, "adding camera properties to camera")
    bpy.types.Camera.noxData = bpy.props.PointerProperty(type=NOXCameraProps)
    
    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXCameraPanel.bl_label)
    bpy.utils.register_class(NOXCameraPanel)
    
    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXRenderPanel.bl_label)
    bpy.utils.register_class(NOXRenderPanel)
    
    NOX_debug.addConsoleLog(logON, "registering blend layer properties")
    bpy.utils.register_class(NOXBlendLayersProps)    

    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXBlendLayersPanel.bl_label)
    bpy.utils.register_class(NOXBlendLayersPanel)

    NOX_debug.addConsoleLog(logON, "registering scene properties")
    bpy.utils.register_class(NOXSceneProps)    

    NOX_debug.addConsoleLog(logON, "adding scene properties to scene")
    bpy.types.Scene.noxData = bpy.props.PointerProperty(type=NOXSceneProps)

    NOX_debug.addConsoleLog(logON, "registering render engine class: " + NOXRenderEngine.bl_label)
    bpy.utils.register_class(NOXRenderEngine)
    
    NOX_debug.addConsoleLog(logON, "registering texture slot properties")
    bpy.utils.register_class(NOXTextureSlotProps)
    
    NOX_debug.addConsoleLog(logON, "registering material layer properties")
    bpy.utils.register_class(NOXMatLayerProps)
    
    NOX_debug.addConsoleLog(logON, "registering material properties")
    bpy.utils.register_class(NOXMaterialProps)
    
    NOX_debug.addConsoleLog(logON, "adding material properties to material")
    bpy.types.Material.noxData = bpy.props.PointerProperty(type=NOXMaterialProps)
    
    NOX_debug.addConsoleLog(logON, "registering panel: " + TestMatPanel.bl_label)
    bpy.utils.register_class(TestMatPanel)
    
    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXTexturePanel.bl_label)
    bpy.utils.register_class(NOXTexturePanel)
    
    NOX_debug.addConsoleLog(logON, "registering sunsky properties")
    bpy.utils.register_class(NOXSunskyProps)

    NOX_debug.addConsoleLog(logON, "adding sunsky properties to world")
    bpy.types.World.noxData = bpy.props.PointerProperty(type=NOXSunskyProps)

    NOX_debug.addConsoleLog(logON, "registering envmap properties")
    bpy.utils.register_class(NOXEnvMapProps)

    NOX_debug.addConsoleLog(logON, "adding envmap properties to world")
    bpy.types.World.noxDataEnv = bpy.props.PointerProperty(type=NOXEnvMapProps)

    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXSunskyPanel.bl_label)
    bpy.utils.register_class(NOXSunskyPanel)

    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXEnvmapPanel.bl_label)
    bpy.utils.register_class(NOXEnvmapPanel)

    NOX_debug.addConsoleLog(logON, "registering operator: " + delete_tex_permanently.bl_label)
    bpy.utils.register_class(delete_tex_permanently)

    NOX_debug.addConsoleLog(logON, "registering operator: " + NOXMaterialImporter.bl_label)
    bpy.utils.register_class(NOXMaterialImporter)
    
    NOX_debug.addConsoleLog(logON, "registering operator: " + NOXMaterialExporter.bl_label)
    bpy.utils.register_class(NOXMaterialExporter)

    NOX_debug.addConsoleLog(logON, "registering operator: " + RunMatEditorOperator.bl_label)
    bpy.utils.register_class(RunMatEditorOperator)
        
    NOX_debug.addConsoleLog(logON, "registering operator: " + ExportNOXScene.bl_label)
    bpy.utils.register_class(ExportNOXScene)
    
    NOX_debug.addConsoleLog(logON, "registering operator: " + RunNOXRendererOperator.bl_label)
    bpy.utils.register_class(RunNOXRendererOperator)
    
    NOX_debug.addConsoleLog(logON, "adding " + ExportNOXScene.bl_label + " to menu")
    bpy.types.INFO_MT_file_export.append(menu_func_export_nox_scene)
    
    NOX_debug.addConsoleLog(logON, "adding " + NOXRenderEngine.bl_label + " engine to materials list panel")
    bpy.types.MATERIAL_PT_context_material.COMPAT_ENGINES.add("nox_renderer")    

    NOX_debug.addConsoleLog(logON, "registering operator: " + eval_sun_pos.bl_label)
    bpy.utils.register_class(eval_sun_pos)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + change_camera_length.bl_label)
    bpy.utils.register_class(change_camera_length)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + set_camera_active.bl_label)
    bpy.utils.register_class(set_camera_active)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + add_nox_image_texture.bl_label)
    bpy.utils.register_class(add_nox_image_texture)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + change_tex_type_to_image.bl_label)
    bpy.utils.register_class(change_tex_type_to_image)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + uncheck_tex_in_nox_slot.bl_label)
    bpy.utils.register_class(uncheck_tex_in_nox_slot)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + add_remove_material_layer.bl_label)
    bpy.utils.register_class(add_remove_material_layer)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + DrawSelectedBoundsOperator.bl_label)
    bpy.utils.register_class(DrawSelectedBoundsOperator)    

    NOX_debug.addConsoleLog(logON, "registering operator: " + DrawSelectedSolidOperator.bl_label)
    bpy.utils.register_class(DrawSelectedSolidOperator)    

    NOX_debug.addConsoleLog(logON, "registering panel: " + NOXToolsPanel.bl_label)
    bpy.utils.register_class(NOXToolsPanel)
 
    NOX_debug.addConsoleLog(logON, "NOX Renderer add-on started - DONE")
    NOX_debug.addConsoleLog(logON, "----------------------------------------------------------")
    
def unregister():
    logON = NOX_debug.log_start_stop_nox_addon

    NOX_debug.addConsoleLog(logON, "----------------------------------------------------------")
    NOX_debug.addConsoleLog(logON, "Stopping NOX Renderer add-on")
    
    NOX_debug.addConsoleLog(logON, "unregistering camera properties")
    bpy.utils.unregister_class(NOXCameraProps)
    
    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXCameraPanel.bl_label)
    bpy.utils.unregister_class(NOXCameraPanel)
    
    NOX_debug.addConsoleLog(logON, "unregistering blend layer properties")
    bpy.utils.unregister_class(NOXBlendLayersProps)    

    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXBlendLayersPanel.bl_label)
    bpy.utils.unregister_class(NOXBlendLayersPanel)

    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXRenderPanel.bl_label)
    bpy.utils.unregister_class(NOXRenderPanel)
    
    NOX_debug.addConsoleLog(logON, "unregistering scene properties")
    bpy.utils.unregister_class(NOXSceneProps)    

    NOX_debug.addConsoleLog(logON, "unregistering sunsky properties")
    bpy.utils.unregister_class(NOXSunskyProps)
    
    NOX_debug.addConsoleLog(logON, "unregistering envmap properties")
    bpy.utils.unregister_class(NOXEnvMapProps)

    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXSunskyPanel.bl_label)
    bpy.utils.unregister_class(NOXSunskyPanel)
    
    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXEnvmapPanel.bl_label)
    bpy.utils.unregister_class(NOXEnvmapPanel)
    
    NOX_debug.addConsoleLog(logON, "unregistering rendering engine:" + NOXRenderEngine.bl_label)
    bpy.utils.unregister_class(NOXRenderEngine)

    NOX_debug.addConsoleLog(logON, "unregistering material properties")
    bpy.utils.unregister_class(NOXMaterialProps)

    NOX_debug.addConsoleLog(logON, "unregistering material layer properties")
    bpy.utils.unregister_class(NOXMatLayerProps)

    NOX_debug.addConsoleLog(logON, "unregistering texture slot properties")
    bpy.utils.unregister_class(NOXTextureSlotProps)
    
    NOX_debug.addConsoleLog(logON, "unregistering panel: " + TestMatPanel.bl_label)
    bpy.utils.unregister_class(TestMatPanel)

    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXTexturePanel.bl_label)
    bpy.utils.unregister_class(NOXTexturePanel)

    NOX_debug.addConsoleLog(logON, "unregistering  operator: " + delete_tex_permanently.bl_label)
    bpy.utils.unregister_class(delete_tex_permanently)

    NOX_debug.addConsoleLog(logON, "unregistering  operator: " + NOXMaterialImporter.bl_label)
    bpy.utils.unregister_class(NOXMaterialImporter)

    NOX_debug.addConsoleLog(logON, "unregistering  operator: " + RunMatEditorOperator.bl_label)
    bpy.utils.unregister_class(RunMatEditorOperator)

    NOX_debug.addConsoleLog(logON, "unregistering  operator: " + NOXMaterialExporter.bl_label)
    bpy.utils.unregister_class(NOXMaterialExporter)

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + ExportNOXScene.bl_label)
    bpy.utils.unregister_class(ExportNOXScene)

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + RunNOXRendererOperator.bl_label)
    bpy.utils.unregister_class(RunNOXRendererOperator)

    NOX_debug.addConsoleLog(logON, "removing " + ExportNOXScene.bl_label + " from menu")
    bpy.types.INFO_MT_file_export.remove(menu_func_export_nox_scene)

    NOX_debug.addConsoleLog(logON, "removing " + NOXRenderEngine.bl_label + " engine from materials list panel")
    bpy.types.MATERIAL_PT_context_material.COMPAT_ENGINES.remove("nox_renderer")    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + eval_sun_pos.bl_label)
    bpy.utils.unregister_class(eval_sun_pos)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + change_camera_length.bl_label)
    bpy.utils.unregister_class(change_camera_length)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + set_camera_active.bl_label)
    bpy.utils.unregister_class(set_camera_active)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + add_nox_image_texture.bl_label)
    bpy.utils.unregister_class(add_nox_image_texture)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + change_tex_type_to_image.bl_label)
    bpy.utils.unregister_class(change_tex_type_to_image)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + uncheck_tex_in_nox_slot.bl_label)
    bpy.utils.unregister_class(uncheck_tex_in_nox_slot)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + add_remove_material_layer.bl_label)
    bpy.utils.unregister_class(add_remove_material_layer)    
    
    NOX_debug.addConsoleLog(logON, "unregistering panel: " + NOXToolsPanel.bl_label)
    bpy.utils.unregister_class(NOXToolsPanel)

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + DrawSelectedBoundsOperator.bl_label)
    bpy.utils.unregister_class(DrawSelectedBoundsOperator)    

    NOX_debug.addConsoleLog(logON, "unregistering operator: " + DrawSelectedSolidOperator.bl_label)
    bpy.utils.unregister_class(DrawSelectedSolidOperator)    
    
    NOX_debug.addConsoleLog(logON, "unregistering message box operator")
    bpy.utils.unregister_class(MessageOperator)
    bpy.utils.unregister_class(OkOperator)

    NOX_debug.addConsoleLog(logON, "NOX Renderer add-on stopped - DONE")
    NOX_debug.addConsoleLog(logON, "----------------------------------------------------------")

if __name__ == "__main__":
    register()

def menu_func_export_nox_scene(self, context):
    self.layout.operator(ExportNOXScene.bl_idname, text="NOX Export (*.nox)")


#---------------------------------------------------------------------------------------------------
