import bpy
#from io_utils import ExportHelper
from bpy_extras.io_utils import ExportHelper
from bpy.props import StringProperty, BoolProperty, EnumProperty
from math import degrees, tan, atan
from mathutils import Vector
import math
import struct
import os
import winreg
from subprocess import call, Popen
from render_nox.debug import NOX_debug
from render_nox.material.material_ops import runNOX

#---------------------------------------------------------------------------------------------------
# global settings

nox_debug = True

nox_version_major = 0
nox_version_minor = 3
nox_version_build = 38

single_tri_size = 106

glb_emitter_exists = False
glb_scene_size = 0
glb_geom_size = 0
glb_num_meshes = 0
glb_num_tris = 0
glb_mat_override = -1
glb_scene = None

#---------------------------------------------------------------------------------------------------

class ExportNOXScene(bpy.types.Operator, ExportHelper):
    '''This appiers in the tooltip of the operator and in the generated docs.'''
    bl_idname = "export.nox_data"
    bl_label = "Export NOX Data"
    filename_ext = ".nox"
    filter_glob = StringProperty(default="*.nox", options={'HIDDEN'})

    #use_setting = BoolProperty(name="Example Boolean", description="Example Tooltip", default=True)
    action_after = EnumProperty(items=(('OPT_A', "Export only", "Export scene to file and do nothing more"),
                                ('OPT_B', "Export and load", "Export scene to file and load it in NOX Renderer"),
                                ('OPT_C', "Export and render", "Export scene to file, load it in NOX Renderer and start rendering"),),
                                name="Action",
                                description="Choose action after export",
                                default='OPT_A')

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        return write_NOX_data(context, self.filepath, self.action_after)

#---------------------------------------------------------------------------------------------------


def write_NOX_data(context, filepath, action):
    global glb_emitter_exists
    global glb_scene
    logON = NOX_debug.log_export_nox_scene

    NOX_debug.addConsoleLog(logON, "running write_NOX_data...")

    write_NOX_scene(filepath, context)

    NOX_debug.addConsoleLog(logON and action=="OPT_A", "DO NOTHING MORE")
    NOX_debug.addConsoleLog(logON and action=="OPT_B", "LOAD SCENE")
    NOX_debug.addConsoleLog(logON and action=="OPT_C", "RENDER SCENE")
    
    go_on = True
    if glb_scene == None:
        go_on = False
    else:
        if not glb_emitter_exists:
            if not glb_scene.noxData.ignore_no_emitter:
                go_on = False

    if go_on:
        if action=="OPT_B":
            runNOX(filepath, False)

        if action=="OPT_C":
            runNOX(filepath, True)

    return {'FINISHED'}

#---------------------------------------------------------------------------------------------------

NOX_BIN_HEADER = 0x20584F4E
NOX_BIN_SCENE = 0x4000

NOX_BIN_PRELOAD_POST = 0x4021
NOX_BIN_PRELOAD_BLEND = 0x4022
NOX_BIN_PRELOAD_FINAL = 0x4023

NOX_BIN_GEOMETRY = 0x4100
NOX_BIN_MESH = 0x4110
NOX_BIN_MESHNAME = 0x4111
NOX_BIN_TRIANGLE = 0x4120
NOX_BIN_SCALE_SCENE = 0x4101
NOX_BIN_INSTANCE_SOURCE = 0x4150
NOX_BIN_INSTANCE = 0x4160
NOX_BIN_INSTANCE_NAME = 0x4161
NOX_BIN_INSTANCE_MATRIX = 0x4162
NOX_BIN_INSTANCE_MATERIALS = 0x4163

NOX_BIN_SCENE_INFO = 0x4020
NOX_BIN_SCENEINFO_SCENENAME = 0x4021
NOX_BIN_SCENEINFO_AUTHOR = 0x4022
NOX_BIN_SCENEINFO_EMAIL = 0x4023
NOX_BIN_SCENEINFO_WWW = 0x4024

NOX_BIN_ENVIRONMENT_MAP = 0x4A01
NOX_BIN_ENV_ENABLED = 0x4A02
NOX_BIN_ENV_POWER_EV = 0x4A04
NOX_BIN_ENV_BLEND_LAYER = 0x4A05
NOX_BIN_ENV_AZIMUTH = 0x4A06
NOX_BIN_ENV_DIRECTIONS = 0x4A07

NOX_BIN_CAMERA = 0x4200
NOX_BIN_CAM_ACTIVE = 0x4201
NOX_BIN_CAM_WIDTH = 0x4202
NOX_BIN_CAM_HEIGHT = 0x4203
NOX_BIN_CAM_AA = 0x4204
NOX_BIN_CAM_NAME = 0x4205
NOX_BIN_CAM_POS = 0x4206
NOX_BIN_CAM_DIR = 0x4207
NOX_BIN_CAM_UPDIR = 0x4208
NOX_BIN_CAM_ANGLE = 0x4209
NOX_BIN_CAM_ISO = 0x420A
NOX_BIN_CAM_APERTURE = 0x420B
NOX_BIN_CAM_SHUTTER = 0x420C
NOX_BIN_CAM_AUTOEXPOSURE_ON = 0x420D
NOX_BIN_CAM_AUTOEXPOSURE_TYPE = 0x420E
NOX_BIN_CAM_AUTOFOCUS = 0x420F
NOX_BIN_CAM_FOCUSDIST = 0x4210
NOX_BIN_CAM_BLADESNUMBER = 0x4211
NOX_BIN_CAM_BLADESANGLE = 0x4212
NOX_BIN_CAM_BLADESROUND = 0x4213
NOX_BIN_CAM_BLADESRADIUS = 0x4214
NOX_BIN_CAM_DOF_ON = 0x4215
NOX_BIN_CAM_BOKEH_RING_SIZE = 0x4216
NOX_BIN_CAM_BOKEH_RING_BALANCE = 0x4217
NOX_BIN_CAM_BOKEH_FLATTEN = 0x4218
NOX_BIN_CAM_BOKEH_VIGNETTE = 0x4219
NOX_BIN_CAM_BUFFERS = 0x4230
NOX_BIN_CAM_BUFFER = 0x4231
NOX_BIN_CAM_DEPTH_BUF = 0x4232

NOX_BIN_MATERIALS = 0x4300
NOX_BIN_MATERIAL = 0x4301
NOX_BIN_MATNAME = 0x4302
NOX_BIN_MAT_PREVIEW = 0x4303
NOX_BIN_MAT_OPACITY = 0x4304
NOX_BIN_MAT_USE_TEX_OPACITY = 0x4305
NOX_BIN_MAT_TEX_OPACITY = 0x4306
NOX_BIN_MAT_SKYPORTAL = 0x4307
NOX_BIN_MATLAYER = 0x4310
NOX_BIN_MLAY_FRESNEL = 0x4311
NOX_BIN_MLAY_FRES_IOR = 0x4312
NOX_BIN_MLAY_REFL0 = 0x4330
NOX_BIN_MLAY_REFL90 = 0x4331
NOX_BIN_MLAY_ROUGH = 0x4332
NOX_BIN_MLAY_TRANSM_ON = 0x4333
NOX_BIN_MLAY_FAKEGLASS_ON = 0x4334
NOX_BIN_MLAY_ABS_ON = 0x4335
NOX_BIN_MLAY_ABS_COLOR = 0x4336
NOX_BIN_MLAY_ABS_DIST = 0x4337
NOX_BIN_MLAY_DISP_ON = 0x434A
NOX_BIN_MLAY_DISP_VALUE = 0x434B
NOX_BIN_MLAY_EMISS_COLOR = 0x4338
NOX_BIN_MLAY_EMISS_TEMP = 0x4339
NOX_BIN_MLAY_EMISS_POWER = 0x433A
NOX_BIN_MLAY_EMISS_UNIT = 0x433B
NOX_BIN_MLAY_EMISS_IES = 0x433C
NOX_BIN_MLAY_REFL0_TEX = 0x4340
NOX_BIN_MLAY_REFL0_TEX_ON = 0x4341
NOX_BIN_MLAY_REFL90_TEX = 0x4342
NOX_BIN_MLAY_REFL90_TEX_ON = 0x4343
NOX_BIN_MLAY_ROUGH_TEX = 0x4344
NOX_BIN_MLAY_ROUGH_TEX_ON = 0x4345
NOX_BIN_MLAY_WEIGHT_TEX = 0x4346
NOX_BIN_MLAY_WEIGHT_TEX_ON = 0x4347
NOX_BIN_MLAY_LIGHT_TEX = 0x4348
NOX_BIN_MLAY_LIGHT_TEX_ON = 0x4349
NOX_BIN_MLAY_NORMAL_TEX = 0x434C
NOX_BIN_MLAY_NORMAL_TEX_ON = 0x434D
NOX_BIN_MLAY_TEXTURE = 0x4370
NOX_BIN_MLAY_TEX_ENABLED = 0x4371
NOX_BIN_MLAY_TEX_FILENAME = 0x4372
NOX_BIN_MLAY_TEX_FILTER = 0x4373
NOX_BIN_MLAY_TEX_PRECALCULATE = 0x4374
NOX_BIN_MLAY_TEX_POST = 0x4375
NOX_BIN_MLAY_TEX_POST_BRIGHTNESS = 0x4376
NOX_BIN_MLAY_TEX_POST_CONTRAST = 0x4377
NOX_BIN_MLAY_TEX_POST_SATURATION = 0x4378
NOX_BIN_MLAY_TEX_POST_HUE = 0x4379
NOX_BIN_MLAY_TEX_POST_RED = 0x437A
NOX_BIN_MLAY_TEX_POST_GREEN = 0x437B
NOX_BIN_MLAY_TEX_POST_BLUE = 0x437C
NOX_BIN_MLAY_TEX_POST_GAMMA = 0x437D
NOX_BIN_MLAY_TEX_POST_INVERT = 0x437E

NOX_BIN_SUNSKY = 0x4400
NOX_BIN_SS_ON = 0x4401
NOX_BIN_SS_MONTH = 0x4402
NOX_BIN_SS_DAY = 0x4403
NOX_BIN_SS_HOUR = 0x4404
NOX_BIN_SS_MINUTE = 0x4405
NOX_BIN_SS_GMT = 0x4406
NOX_BIN_SS_LONGITUDE = 0x4407
NOX_BIN_SS_LATITUDE = 0x4408
NOX_BIN_SS_TURBIDITY = 0x4409
NOX_BIN_SS_AEROSOL = 0x440A
NOX_BIN_SS_SUN_SIZE = 0x440B
NOX_BIN_SS_SUN_ON = 0x440C
NOX_BIN_SS_SUN_OTHER_LAYER = 0x440D
NOX_BIN_SS_DIRECTIONS = 0x440E

NOX_BIN_BLEND_LAYERS = 0x4500
NOX_BIN_BLEND_LAYER = 0x4510
NOX_BIN_BLEND_LAYER_NAME = 0x4511
NOX_BIN_BLEND_LAYER_ON = 0x4512
NOX_BIN_BLEND_LAYER_WEIGHT = 0x4513
NOX_BIN_BLEND_LAYER_RED = 0x4514
NOX_BIN_BLEND_LAYER_GREEN = 0x4515
NOX_BIN_BLEND_LAYER_BLUE = 0x4516

NOX_BIN_USERCOLORS = 0x4550
NOX_BIN_USER_SINGLE_COLOR = 0x4551

NOX_BIN_CAM_POST = 0x4250
NOX_BIN_CAM_POST_ISO = 0x4251
NOX_BIN_CAM_POST_EV = 0x4263
NOX_BIN_CAM_POST_TONE = 0x4252
NOX_BIN_CAM_POST_GAMMA = 0x425A
NOX_BIN_CAM_POST_BRIGHTNESS = 0x425B
NOX_BIN_CAM_POST_CONTRAST = 0x425C
NOX_BIN_CAM_POST_SATURATION = 0x425D
NOX_BIN_CAM_POST_ADD_RED = 0x425E
NOX_BIN_CAM_POST_ADD_GREEN = 0x425F
NOX_BIN_CAM_POST_ADD_BLUE = 0x4260
NOX_BIN_CAM_POST_TEMPERATURE = 0x4261
NOX_BIN_CAM_POST_VIGNETTE = 0x4280
NOX_BIN_CAM_POST_RESPONSE_FUNCTION = 0x4281
NOX_BIN_CAM_POST_FILTER = 0x4262
NOX_BIN_CAM_POST_CURVE = 0x426A
NOX_BIN_CAM_POST_CURVE_PR = 0x426B
NOX_BIN_CAM_POST_CURVE_PG = 0x426C
NOX_BIN_CAM_POST_CURVE_PB = 0x426D
NOX_BIN_CAM_POST_CURVE_PL = 0x426E
NOX_BIN_CAM_POST_BLOOM = 0x4270
NOX_BIN_CAM_POST_BLOOM_ON = 0x4271
NOX_BIN_CAM_POST_BLOOM_AREA = 0x4272
NOX_BIN_CAM_POST_BLOOM_ATTENUATION_ON = 0x4273
NOX_BIN_CAM_POST_BLOOM_ATTENUATION_COLOR = 0x4274
NOX_BIN_CAM_POST_BLOOM_THRESHOLD = 0x4275
NOX_BIN_CAM_POST_GLARE = 0x4278
NOX_BIN_CAM_POST_GLARE_ON = 0x4279
NOX_BIN_CAM_POST_GLARE_DISPERSION = 0x427A
NOX_BIN_CAM_POST_GLARE_TEX_DIAPHRAGM = 0x427B
NOX_BIN_CAM_POST_GLARE_TEX_DIAPHRAGM_ON = 0x427C
NOX_BIN_CAM_POST_GLARE_TEX_OBSTACLE = 0x427D
NOX_BIN_CAM_POST_GLARE_TEX_OBSTACLE_ON = 0x427E
NOX_BIN_CAM_POST_DOTS_ENABLED = 0x4282
NOX_BIN_CAM_POST_GI_EV = 0x4283

NOX_BIN_FINAL_POST = 0x42A0
NOX_BIN_FINAL_QUALITY = 0x42A1
NOX_BIN_FINAL_THREADS = 0x42A2
NOX_BIN_FINAL_ALGORITHM = 0x42A3
NOX_BIN_FINAL_FAKE_DOF_ON = 0x42A8
NOX_BIN_FINAL_FOCUS_DISTANCE = 0x42A9
NOX_BIN_FINAL_APERTURE = 0x42B0
NOX_BIN_FINAL_DIAPH_NUM_BLADES = 0x42B1
NOX_BIN_FINAL_DIAPH_ANGLE = 0x42B2
NOX_BIN_FINAL_DIAPH_RADIUS = 0x42B3
NOX_BIN_FINAL_DIAPH_BLADES_ROUND = 0x42B4
NOX_BIN_FINAL_BOKEH_RING_BALANCE = 0x42B5
NOX_BIN_FINAL_BOKEH_RING_SIZE = 0x42B6
NOX_BIN_FINAL_BOKEH_FLATTEN = 0x42B7
NOX_BIN_FINAL_BOKEH_VIGNETTE = 0x42B8
NOX_BIN_FINAL_ABERRATION_SHIFT_LENS = 0x42BA
NOX_BIN_FINAL_ABERRATION_SHIFT_DEPTH = 0x42BB
NOX_BIN_FINAL_ABERRATION_ACHROMATIC = 0x42BC
NOX_BIN_FINAL_BLOOM_ON = 0x42C0
NOX_BIN_FINAL_BLOOM_POWER = 0x42C1
NOX_BIN_FINAL_BLOOM_AREA = 0x42C2
NOX_BIN_FINAL_BLOOM_THRESHOLD = 0x42C3
NOX_BIN_FINAL_BLOOM_ATTENUATION_ON = 0x42C4
NOX_BIN_FINAL_BLOOM_ATTENUATION_COLOR = 0x42C5
NOX_BIN_FINAL_GLARE_ON = 0x42C7
NOX_BIN_FINAL_GLARE_POWER = 0x42C8
NOX_BIN_FINAL_GLARE_DISPERSION = 0x42C9
NOX_BIN_FINAL_GLARE_THRESHOLD = 0x42CA
NOX_BIN_FINAL_GLARE_SHAPE_ON = 0x42CB
NOX_BIN_FINAL_GLARE_SHAPE_TEX = 0x42CC
NOX_BIN_FINAL_GLARE_OBSTACLE_ON = 0x42CD
NOX_BIN_FINAL_GLARE_OBSTACLE_TEX = 0x42CE
NOX_BIN_FINAL_GLARE_AREA = 0x42CF
NOX_BIN_FINAL_DOTS_ON = 0x42D2
NOX_BIN_FINAL_DOTS_RADIUS = 0x42D3
NOX_BIN_FINAL_DOTS_THRESHOLD = 0x42D4
NOX_BIN_FINAL_DOTS_DENSITY = 0x42D5
NOX_BIN_FINAL_GRAIN = 0x42D6
NOX_BIN_FINAL_VIGNETTE = 0x42D7
NOX_BIN_FINAL_ABERRATION_RED = 0x42D8
NOX_BIN_FINAL_ABERRATION_GREEN = 0x42D9
NOX_BIN_FINAL_ABERRATION_BLUE = 0x42DA

NOX_BIN_RENDERER_SETTINGS = 0x4600
NOX_BIN_RSETT_TIMER_ON_STOPAFTER = 0x4601
NOX_BIN_RSETT_TIMER_ON_REFRESH = 0x4602
NOX_BIN_RSETT_TIMER_ON_AUTOSAVE = 0x4603
NOX_BIN_RSETT_TIMER_STOPAFTER_HOURS = 0x4607
NOX_BIN_RSETT_TIMER_STOPAFTER_MINUTES = 0x4608
NOX_BIN_RSETT_TIMER_STOPAFTER_SECONDS = 0x4609
NOX_BIN_RSETT_TIMER_REFRESH_SECONDS = 0x460A
NOX_BIN_RSETT_TIMER_AUTOSAVE_MINUTES = 0x460B

LAYER_TEX_SLOT_COLOR0 = 0
LAYER_TEX_SLOT_COLOR90 = 1
LAYER_TEX_SLOT_ROUGHNESS = 2
LAYER_TEX_SLOT_WEIGHT = 3
LAYER_TEX_SLOT_LIGHT = 4
LAYER_TEX_SLOT_NORMAL = 5
MAT_TEX_SLOT_OPACITY = 100
TEX_SLOT_ENVMAP = 200

LAYER_TYPE_EMITTER = 1
LAYER_TYPE_SHADE = 2

EMISSION_UNIT_WATT = 1
EMISSION_UNIT_WATT_PER_SQR_METER = 2
EMISSION_UNIT_LUMEN = 3
EMISSION_UNIT_LUX = 4


#---------------------------------------------------------------------------------------------------

def eval_instance_source_size(sce, mesh):
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "Checking instance source "+mesh.name)

    mesh_size = eval_mesh_size(sce, mesh)
    inst_src_size = mesh_size + 14
    return inst_src_size

#---------------------------------------------------------------------------------------------------

def eval_instance_size(sce, meshob):
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "Checking instance "+meshob.name)

    nummats = meshob.material_slots.__len__()
    if nummats == 0:
        nummats = 1
    mats_size = nummats*4+4
    name_size = 6+len(meshob.name)
    inst_size = name_size + mats_size + 14 + 66
    return inst_size

#---------------------------------------------------------------------------------------------------

def eval_mesh_size(sce, mesh):
    logON = NOX_debug.log_export_nox_scene
    global single_tri_size
    global glb_num_tris
    name_size = 6+len(mesh.name)
    mesh_size = 14 + name_size
    total_mesh_tris = 0
    #mesh = meshob.to_mesh(sce, True, "RENDER")
    NOX_debug.addConsoleLog(logON, "Checking mesh "+mesh.name)
    #for face in mesh.faces:
    for face in mesh.polygons:
        ntris = face.vertices.__len__()-2
        total_mesh_tris += ntris
        #print("face: "+str(ntris))
    NOX_debug.addConsoleLog(logON, "Total tris in mesh: "+str(total_mesh_tris))
    #if mesh.name in bpy.data.meshes:
    #    bpy.data.meshes.remove(mesh)
    mesh_size += total_mesh_tris*single_tri_size
    glb_num_tris += total_mesh_tris
    return mesh_size

#---------------------------------------------------------------------------------------------------

#def eval_mesh_size(sce, meshob):
#    logON = NOX_debug.log_export_nox_scene
#    global single_tri_size
#    global glb_num_tris
#    name_size = 6+len(meshob.name)
#    mesh_size = 14 + name_size
#    total_mesh_tris = 0
#    mesh = meshob.to_mesh(sce, True, "RENDER")
#    NOX_debug.addConsoleLog(logON, "Checking mesh "+meshob.name)
#    for face in mesh.faces:
#        ntris = face.vertices.__len__()-2
#        total_mesh_tris += ntris
#        #print("face: "+str(ntris))
#    NOX_debug.addConsoleLog(logON, "Total tris in mesh: "+str(total_mesh_tris))
#    if mesh.name in bpy.data.meshes:
#        bpy.data.meshes.remove(mesh)
#    mesh_size += total_mesh_tris*single_tri_size
#    glb_num_tris += total_mesh_tris
#    return mesh_size

#---------------------------------------------------------------------------------------------------

def eval_geom_size(sce):
    global glb_geom_size
    global glb_num_meshes
    geom_size = 22
    
    for ob in sce.objects:
        #print("object type: "+ob.type)
        if (ob.type=="MESH"):
            if is_object_visible(sce, ob):
                #print("got mesh")
                glb_num_meshes += 1
                #mesh_size = eval_mesh_size(sce, ob)
                inst_size = eval_instance_size(sce, ob)
                geom_size += inst_size

    for mesh in bpy.data.meshes:
        use_it = False
        for ob in sce.objects:
            if (ob.type=="MESH"):
                if is_object_visible(sce, ob):
                    if (ob.data.id_data.name==mesh.name):
                        use_it = True
        if use_it:
            geom_size += eval_instance_source_size(sce, mesh)
                
    glb_geom_size = geom_size
    return geom_size

#---------------------------------------------------------------------------------------------------

#def eval_geom_size(sce):
#    global glb_geom_size
#    global glb_num_meshes
#    geom_size = 22
#    for ob in sce.objects:
#        #print("object type: "+ob.type)
#        if (ob.type=="MESH"):
#            if is_object_visible(sce, ob):
#                #print("got mesh")
#                glb_num_meshes += 1
#                mesh_size = eval_mesh_size(sce, ob)
#                geom_size += mesh_size
#    glb_geom_size = geom_size
#    return geom_size

#---------------------------------------------------------------------------------------------------

def eval_camera_size(sce, camob):
    cam_size = 0
    if (camob.type=="CAMERA"):
        cam_size = 78 + 64
        nsize = len(camob.name)+6
    return cam_size+nsize

#---------------------------------------------------------------------------------------------------

def eval_cameras_size(sce):
    cams_size = 0
    for ob in sce.objects:
        if (ob.type=="CAMERA"):
            cam_size = eval_camera_size(sce, ob)
            cams_size += cam_size
    return cams_size

#---------------------------------------------------------------------------------------------------

def eval_material_default_size():
    return 196

#---------------------------------------------------------------------------------------------------

def eval_texture_size(mnum, lnum, mslot):
    global glb_scene
    logON = NOX_debug.log_export_nox_scene
    got_slot = False

    if mslot == TEX_SLOT_ENVMAP:
        world = glb_scene.world
        #world = bpy.context.scene.world
        if world:
            if hasattr(world, "noxDataEnv"):
                tslot = world.noxDataEnv.texture
                got_slot = True
    else:
        mat = bpy.data.materials[mnum]
        if not hasattr(mat, "noxData"):
            return 0

        if mslot == MAT_TEX_SLOT_OPACITY:
            tslot = mat.noxData.tex_opacity
            got_slot = True
        else:
            mlay = mat.noxData.layers[lnum]
            if mslot == LAYER_TEX_SLOT_COLOR0:
                tslot = mlay.tex_refl0
                got_slot = True
            if mslot == LAYER_TEX_SLOT_COLOR90:
                tslot = mlay.tex_refl90
                got_slot = True
            if mslot == LAYER_TEX_SLOT_ROUGHNESS:
                tslot = mlay.tex_rough
                got_slot = True
            if mslot == LAYER_TEX_SLOT_WEIGHT:
                tslot = mlay.tex_weight
                got_slot = True
            if mslot == LAYER_TEX_SLOT_LIGHT:
                tslot = mlay.tex_light
                got_slot = True
            if mslot == LAYER_TEX_SLOT_NORMAL:
                tslot = mlay.tex_normal
                got_slot = True

    if not got_slot:
        NOX_debug.addConsoleLog(logON, "Internal Error: Texture slot not available.\nmatid="+str(mnum)+"layerid="+str(lnum)+"slot="+str(mslot))
        return 0
        
    fullfilename = ''
    if tslot.texID<bpy.data.textures.__len__() and tslot.texID>-1:
        tex = bpy.data.textures[tslot.texID]
        if hasattr(tex, "image") and tex.type=="IMAGE":
            if hasattr(tex.image, 'filepath') and hasattr(tex.image, 'source'):
                if tex.image.source=='FILE' and tex.image.filepath!='':
                    fullfilename = tex.image.filepath
    
    if fullfilename!='':
        absfullfilename = bpy.path.abspath(fullfilename)
        fnlen = 6 + len(absfullfilename)
    else:
        fnlen = 0

    return 84 + fnlen

#---------------------------------------------------------------------------------------------------

def eval_ies_size(mnum, lnum):
    logON = NOX_debug.log_export_nox_scene
    mat = bpy.data.materials[mnum]
    if not hasattr(mat, "noxData"):
        return 0
    mlay = mat.noxData.layers[lnum]
    if mlay.ies_file.__len__()<1:
        return 0
    absfullfilename = bpy.path.abspath(mlay.ies_file)
    return 6 + len(absfullfilename)

#---------------------------------------------------------------------------------------------------

def eval_material_size(mnum):
    mat = bpy.data.materials[mnum]
    name_size = 6+len(mat.name)
    return 180+name_size

#---------------------------------------------------------------------------------------------------

def eval_material_layer_nox_size(mnum, lnum):
    stex_col0 = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_COLOR0)
    stex_col90 = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_COLOR90)
    stex_rough = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_ROUGHNESS)
    stex_weight = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_WEIGHT)
    stex_light = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_LIGHT)
    stex_normal = eval_texture_size(mnum, lnum, LAYER_TEX_SLOT_NORMAL)

    s_ies = eval_ies_size(mnum, lnum)

    return 152 + stex_col0 + stex_col90 + stex_rough + stex_weight + stex_light + stex_normal + s_ies
    
#---------------------------------------------------------------------------------------------------

def eval_material_nox_size(mnum):
    mat = bpy.data.materials[mnum]
    name_size = 6+len(mat.name)
    #return 180+name_size
    layers_size = 0
    
    stex_opacity = eval_texture_size(mnum, 0, MAT_TEX_SLOT_OPACITY)
    for i in range(mat.noxData.layers.__len__()):
        layer_size = eval_material_layer_nox_size(mnum, i)
        layers_size += layer_size
    return 28 + 4 + name_size + layers_size + stex_opacity

#---------------------------------------------------------------------------------------------------

def eval_materials_size(sce):
    global glb_mat_override
    total_mats_size = 14

    if glb_mat_override>-1:
        gmat = bpy.data.materials[glb_mat_override]
        if hasattr(gmat, "noxData"):
            if gmat.noxData.layers.__len__() > 0:
                defsize = eval_material_nox_size(glb_mat_override)
            else:
                defsize = eval_material_size(glb_mat_override)
        else:
            defsize = eval_material_size(glb_mat_override)
    else:
        defsize = eval_material_default_size()
    
    num_mats = bpy.data.materials.__len__()
    for i in range(num_mats):
        mat = bpy.data.materials[i]
        use_this = True
        if glb_mat_override > -1:
            if not is_nox_emitter(i):
                use_this = False
        if use_this:
            if hasattr(mat, "noxData"):
                if mat.noxData.layers.__len__() > 0:
                    total_mats_size += eval_material_nox_size(i)
                else:
                    total_mats_size += eval_material_size(i)
            else:
                total_mats_size += eval_material_size(i)
    total_mats_size += defsize

    return total_mats_size

#---------------------------------------------------------------------------------------------------

def eval_blend_layers_size(sce):
    bl_size = 0
    if not hasattr(sce, "noxData"):
        return 0
    allstrsize = 0
    allstrsize += len(sce.noxData.blend_layers.b1)
    allstrsize += len(sce.noxData.blend_layers.b2)
    allstrsize += len(sce.noxData.blend_layers.b3)
    allstrsize += len(sce.noxData.blend_layers.b4)
    allstrsize += len(sce.noxData.blend_layers.b5)
    allstrsize += len(sce.noxData.blend_layers.b6)
    allstrsize += len(sce.noxData.blend_layers.b7)
    allstrsize += len(sce.noxData.blend_layers.b8)
    allstrsize += len(sce.noxData.blend_layers.b9)
    allstrsize += len(sce.noxData.blend_layers.b10)
    allstrsize += len(sce.noxData.blend_layers.b11)
    allstrsize += len(sce.noxData.blend_layers.b12)
    allstrsize += len(sce.noxData.blend_layers.b13)
    allstrsize += len(sce.noxData.blend_layers.b14)
    allstrsize += len(sce.noxData.blend_layers.b15)
    allstrsize += len(sce.noxData.blend_layers.b16)
    
    bl_size = allstrsize + 10 + 16*14 + 16*6
    return bl_size

#---------------------------------------------------------------------------------------------------

def eval_sunsky_size(sce):
    world = sce.world
    if world:
        if hasattr(world, "noxData"):
            return 98
    return 0

#---------------------------------------------------------------------------------------------------

def eval_envmap_size(sce):
    world = sce.world
    if world:
        if hasattr(world, "noxDataEnv"):
            tex_size = eval_texture_size(0, 0, TEX_SLOT_ENVMAP)
            return 56 + tex_size
    return 0

#---------------------------------------------------------------------------------------------------

def eval_preload_stuff_size(sce):
    result = 0
    if hasattr(sce, "noxData"):
        if len(sce.noxData.preload_post_file)>0  and  sce.noxData.preload_post_on:
            fullfilename = sce.noxData.preload_post_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            result += fnlen
        if len(sce.noxData.preload_blend_file)>0  and  sce.noxData.preload_blend_on:
            fullfilename = sce.noxData.preload_blend_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            result += fnlen
        if len(sce.noxData.preload_final_file)>0  and  sce.noxData.preload_final_on:
            fullfilename = sce.noxData.preload_final_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            result += fnlen
    return result
    
#---------------------------------------------------------------------------------------------------

def eval_scene_size(sce):
    global glb_scene_size
    scene_size = 10

    geom_size = eval_geom_size(sce)
    cams_size = eval_cameras_size(sce)
    mats_size = eval_materials_size(sce)
    bl_size = eval_blend_layers_size(sce)
    sunsky_size = eval_sunsky_size(sce)
    env_size = eval_envmap_size(sce)
    preload_size = eval_preload_stuff_size(sce)

    scene_size += geom_size
    scene_size += cams_size
    scene_size += mats_size
    scene_size += bl_size
    scene_size += sunsky_size
    scene_size += env_size
    scene_size += preload_size

    glb_scene_size = scene_size
    return scene_size

#---------------------------------------------------------------------------------------------------

def add_mesh(sce, file, mesh):
    global single_tri_size
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "Exporting mesh "+mesh.name)
        
    #mesh = meshob.to_mesh(sce, True, "RENDER")
    #mesh = meshob.data.id_data
    #mesh.transform(meshob.matrix_world)
    mesh.calc_normals()	

    meshob = None
    for ob in sce.objects:
        if (ob.type=="MESH"):
            if is_object_visible(sce, ob):
                if (ob.data.id_data.name==mesh.name):
                    meshob = ob
    if meshob is None:
        NOX_debug.addConsoleLog(logON, "Exporting mesh error - can't find object with mesh named "+mesh.name)
        return

    mats = []
    for msl in meshob.material_slots:
        sl_mat_id = 0
        matname = msl.material.name
        for i in range(bpy.data.materials.__len__()):
            dmat = bpy.data.materials[i]
            if matname==dmat.name:
                sl_mat_id = i+1
        mats.append(sl_mat_id)
    num_slots_used = mats.__len__()

    name_size = 6+len(mesh.name)
    num_mesh_tris = 0
    for face in mesh.polygons:
        num_mesh_tris += face.vertices.__len__()-2
    #for face in mesh.faces:
    #    num_mesh_tris += face.vertices.__len__()-2
		

    mesh_size = 14 + name_size
    mesh_size += num_mesh_tris*single_tri_size

    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_MESH)
    file.write(data)
    data=struct.pack("=Q", mesh_size)
    file.write(data)
    data=struct.pack("=I", num_mesh_tris)
    file.write(data)

    data=struct.pack("=H", NOX_BIN_MESHNAME)
    file.write(data)
    data=struct.pack("=I", name_size)
    file.write(data)

    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode(meshob.name))
    file.write(data)


    got_uv_channel = False
    if mesh.uv_textures.__len__()>0:
        got_uv_channel = True
    uv_u1 = 0
    uv_u2 = 0
    uv_u3 = 0
    uv_v1 = 0
    uv_v2 = 0
    uv_v3 = 0

    num_vertices = mesh.vertices.__len__()
    
    
    # init smooth normals 
    snormals = []
    for i in range(num_vertices):
        snormals.append(Vector([0,0,0]))
    #for j in range(mesh.faces.__len__()):
    #    face = mesh.faces[j]
    for j in range(mesh.polygons.__len__()):
        face = mesh.polygons[j]
        tris = face.vertices.__len__()-2
        if face.use_smooth:
            for i in range(ntris):
                fvert1 = face.vertices[0]
                fvert2 = face.vertices[i+1]
                fvert3 = face.vertices[i+2]
                v1 = mesh.vertices[fvert1].co
                v2 = mesh.vertices[fvert2].co
                v3 = mesh.vertices[fvert3].co
                a1 = getTriAngle(v3,v1,v2)
                a2 = getTriAngle(v1,v2,v3)
                a3 = getTriAngle(v2,v3,v1)
                norm = face.normal
                snormals[fvert1] += norm*a1
                snormals[fvert2] += norm*a2
                snormals[fvert3] += norm*a3
                #print("angles: "+str(a1)+"   "+str(a2)+"   "+str(a3))
        #else:
        #    print("no smooth face")

    for i in range(num_vertices):
        snormals[i].normalize()
        #print(str(i) + " = (" + str(snormals[i][0])+" , " + str(snormals[i][1])+" , " + str(snormals[i][2])+")")
        #print("v = (" + str(mesh.vertices[i].co[0]) + " , " + str(mesh.vertices[i].co[1]) + " , " + str(mesh.vertices[i].co[2]) + ")")


    # add to file
    #for j in range(mesh.faces.__len__()):
    #    face = mesh.faces[j]
    for j in range(mesh.polygons.__len__()):
        face = mesh.polygons[j]
        if got_uv_channel:
            uvface = mesh.uv_textures[0].data[j]
        ntris = face.vertices.__len__()-2
        for i in range(ntris):
            fvert1 = face.vertices[0]
            fvert2 = face.vertices[i+1]
            fvert3 = face.vertices[i+2]
            v1 = mesh.vertices[fvert1].co
            v2 = mesh.vertices[fvert2].co
            v3 = mesh.vertices[fvert3].co
            norm = face.normal
            norm1 = snormals[fvert1]
            norm2 = snormals[fvert2]
            norm3 = snormals[fvert3]
            #norm.normalize()
            if face.use_smooth:
                n1_1 = norm1[0]
                n1_2 = norm1[2]
                n1_3 = -norm1[1]
                n2_1 = norm2[0]
                n2_2 = norm2[2]
                n2_3 = -norm2[1]
                n3_1 = norm3[0]
                n3_2 = norm3[2]
                n3_3 = -norm3[1]
            else:
                n1_1 = norm[0]
                n1_2 = norm[2]
                n1_3 = -norm[1]
                n2_1 = norm[0]
                n2_2 = norm[2]
                n2_3 = -norm[1]
                n3_1 = norm[0]
                n3_2 = norm[2]
                n3_3 = -norm[1]
            
            #print("tri "+str(i))
            #print("v1  " + str(v1[0]) + "    " + str(v1[2]) + "    " + str(-v1[1]))
            #print("n1  " + str(n1_1)  + "    " + str(n1_2)  + "    " + str(n1_3) )
            #print("v2  " + str(v2[0]) + "    " + str(v2[2]) + "    " + str(-v2[1]))
            #print("n2  " + str(n2_1)  + "    " + str(n2_2)  + "    " + str(n2_3) )
            #print("v3  " + str(v3[0]) + "    " + str(v3[2]) + "    " + str(-v3[1]))
            #print("n3  " + str(n3_1)  + "    " + str(n3_2)  + "    " + str(n3_3) )
            
            if got_uv_channel:
                uv_u1 = uvface.uv[0][0]
                uv_u2 = uvface.uv[i+1][0]
                uv_u3 = uvface.uv[i+2][0]
                uv_v1 = uvface.uv[0][1]
                uv_v2 = uvface.uv[i+1][1]
                uv_v3 = uvface.uv[i+2][1]
			
            data=struct.pack("=H", NOX_BIN_TRIANGLE)
            file.write(data)
            data=struct.pack("=I", face.material_index)
            file.write(data)

            matID = 0
            if face.material_index<num_slots_used:
                matID = mats[face.material_index]
            zeroval = 0;
            data=struct.pack("=I", matID)
            file.write(data)
            #verts
            data=struct.pack("=fff", v1[0], v1[2], -v1[1])
            file.write(data)
            data=struct.pack("=fff", v2[0], v2[2], -v2[1])
            file.write(data)
            data=struct.pack("=fff", v3[0], v3[2], -v3[1])
            file.write(data)

            #print("adding vert "+ str(v1[0]) +"  "+ str(v1[1]) +"  "+ str(v1[2]))
            #print("adding vert "+ str(v2[0]) +"  "+ str(v2[1]) +"  "+ str(v2[2]))
            #print("adding vert "+ str(v3[0]) +"  "+ str(v3[1]) +"  "+ str(v3[2]))
            #print("adding norm "+ str(n1) +"  "+ str(n2) +"  "+ str(n3))

            #normals
            data=struct.pack("=fff", n1_1, n1_2, n1_3)
            file.write(data)
            data=struct.pack("=fff", n2_1, n2_2, n2_3)
            file.write(data)
            data=struct.pack("=fff", n3_1, n3_2, n3_3)
            file.write(data)
            #uv
            data=struct.pack("=ff", uv_u1, uv_v1)
            file.write(data)
            data=struct.pack("=ff", uv_u2, uv_v2)
            file.write(data)
            data=struct.pack("=ff", uv_u3, uv_v3)
            file.write(data)

    #if mesh.name in bpy.data.meshes:
    #    bpy.data.meshes.remove(mesh)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != mesh_size):
        NOX_debug.addConsoleLog(logON, "Mesh size incorrect: "+str(diff_pos)+" instead of "+str(mesh_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def getTriAngle(v1,v2,v3):  # angle in v2
    vec1 = v1-v2
    vec2 = v3-v2
    vec1.normalize()
    vec2.normalize()
    cangle = max(-1.0, min(1.0, vec1*vec2))
    #print("angle = "+str(cangle))
    angle = math.acos(cangle)*180/math.pi
    return angle

#---------------------------------------------------------------------------------------------------

def is_object_visible(sce, ob):
    if ob.hide:
        return False
    for i in range(20):
        if ob.layers[i] and sce.layers[i]:
            return True
    return False

#---------------------------------------------------------------------------------------------------

def add_instance_source(sce, file, meshid):
    logON = NOX_debug.log_export_nox_scene
    
    #TODO: check array boundary
    mesh = bpy.data.meshes[meshid]
    inst_size = eval_instance_source_size(sce, mesh)

    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_INSTANCE_SOURCE)
    file.write(data)
    data=struct.pack("=Q", inst_size)
    file.write(data)
    data=struct.pack("=I", meshid)
    file.write(data)

    add_mesh(sce, file, mesh)
    
    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != inst_size):
        NOX_debug.addConsoleLog(logON, "Instance source size incorrect: "+str(diff_pos)+" instead of "+str(inst_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def add_instance(sce, file, meshob):
    global glb_mat_override
    global glb_emitter_exists
    logON = NOX_debug.log_export_nox_scene
    
    inst_size = eval_instance_size(sce, meshob)
    
    meshid = -1
    mcount = bpy.data.meshes.__len__()
    for i in range(mcount):
        mesh = bpy.data.meshes[i]
        if mesh.name==meshob.data.id_data.name:
            meshid = i
    if meshid<0:
        NOX_debug.addConsoleLog(logON, "Error, can't find mesh named "+meshob.data.id_data.name)
        return
        
    mats = []
    for msl in meshob.material_slots:
        sl_mat_id = 0
        if not (msl.material is None):
            matname = msl.material.name
            for i in range(bpy.data.materials.__len__()):
                dmat = bpy.data.materials[i]
                if matname==dmat.name:
                    sl_mat_id = i+1
        if is_nox_emitter(sl_mat_id-1):
            glb_emitter_exists = True
        if glb_mat_override>-1:
            if is_nox_emitter(sl_mat_id-1):
                mats.append(sl_mat_id)
            else:
                mats.append(glb_mat_override+1)
        else:
            mats.append(sl_mat_id)
    num_slots_used = mats.__len__()
    if num_slots_used == 0:
        num_slots_used = 1
        mats.append(0)
    
    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_INSTANCE)
    file.write(data)
    data=struct.pack("=Q", inst_size)
    file.write(data)
    data=struct.pack("=I", meshid)
    file.write(data)

    name_size = len(meshob.name)+6
    data=struct.pack("=HI", NOX_BIN_INSTANCE_NAME, name_size)
    file.write(data) #6
    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode(meshob.name))
    file.write(data) #xxx
    
    data=struct.pack("=H", NOX_BIN_INSTANCE_MATRIX)
    file.write(data)
    data=struct.pack("=ffffffffffffffff", 
                meshob.matrix_world[0][0],   meshob.matrix_world[0][2], -meshob.matrix_world[0][1],  meshob.matrix_world[0][3],
                meshob.matrix_world[2][0],   meshob.matrix_world[2][2], -meshob.matrix_world[2][1],  meshob.matrix_world[2][3],
                -meshob.matrix_world[1][0], -meshob.matrix_world[1][2],  meshob.matrix_world[1][1], -meshob.matrix_world[1][3],
                meshob.matrix_world[3][0],   meshob.matrix_world[3][2], -meshob.matrix_world[3][1],  meshob.matrix_world[3][3])
    #data=struct.pack("=ffffffffffffffff", 
    #            meshob.matrix_world[0][0],   meshob.matrix_world[2][0], -meshob.matrix_world[1][0],  meshob.matrix_world[3][0],
    #            meshob.matrix_world[0][2],   meshob.matrix_world[2][2], -meshob.matrix_world[1][2],  meshob.matrix_world[3][2],
    #            -meshob.matrix_world[0][1], -meshob.matrix_world[2][1],  meshob.matrix_world[1][1], -meshob.matrix_world[3][1],
    #            meshob.matrix_world[0][3],   meshob.matrix_world[2][3], -meshob.matrix_world[1][3],  meshob.matrix_world[3][3])
    file.write(data)
    
    #for i in range(4):
    #    for j in range(4):
    #        data=struct.pack("=f", meshob.matrix_world[j][i])
    #        file.write(data)

    data=struct.pack("=HH", NOX_BIN_INSTANCE_MATERIALS, num_slots_used)
    file.write(data)
    for i in range(num_slots_used):
        data=struct.pack("=I", mats[i])
        file.write(data)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != inst_size):
        NOX_debug.addConsoleLog(logON, "Instance size incorrect: "+str(diff_pos)+" instead of "+str(inst_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def add_geometry(sce, file):
    logON = NOX_debug.log_export_nox_scene
    global glb_geom_size
    global glb_num_meshes
    global glb_num_tris
    global single_tri_size

    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_GEOMETRY)
    file.write(data)
    data=struct.pack("=Q", glb_geom_size)
    file.write(data)
    data=struct.pack("=I", glb_num_meshes)
    file.write(data)
    data=struct.pack("=I", single_tri_size)
    file.write(data)
    data=struct.pack("=I", glb_num_tris)
    file.write(data)

    #for mesh in bpy.data.meshes:
    mcount = bpy.data.meshes.__len__()
    for i in range(mcount):
        mesh = bpy.data.meshes[i]
        use_it = False
        for ob in sce.objects:
            if (ob.type=="MESH"):
                if is_object_visible(sce, ob):
                    if (ob.data.id_data.name==mesh.name):
                        use_it = True
        if use_it:
            add_instance_source(sce, file, i)

    for ob in sce.objects:
        #print("object type: "+ob.type)
        if (ob.type=="MESH"):
            if is_object_visible(sce, ob):
                #add_mesh(sce, file, ob)
                add_instance(sce, file, ob)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != glb_geom_size):
        NOX_debug.addConsoleLog(logON, "Geometry size incorrect: "+str(diff_pos)+" instead of "+str(glb_geom_size)+".\nScene will be corrupted.")


#---------------------------------------------------------------------------------------------------

def add_camera(sce, file, camob):
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "adding camera - " + camob.name)

    cam_size = eval_camera_size(sce, camob)
    width = sce.render.resolution_x
    height = sce.render.resolution_y
    aa = 2
    #mmm = camob.matrix_world
    #print("Camera matrix:")
    #print(str(mmm[0][0])+"  "+str(mmm[0][1])+"  "+str(mmm[0][2])+"  "+str(mmm[0][3]))
    #print(str(mmm[1][0])+"  "+str(mmm[1][1])+"  "+str(mmm[1][2])+"  "+str(mmm[1][3]))
    #print(str(mmm[2][0])+"  "+str(mmm[2][1])+"  "+str(mmm[2][2])+"  "+str(mmm[2][3]))
    #print(str(mmm[3][0])+"  "+str(mmm[3][1])+"  "+str(mmm[3][2])+"  "+str(mmm[3][3]))
    #cam_dir_x = -camob.matrix_world[2][0]
    #cam_dir_y = -camob.matrix_world[2][2]
    #cam_dir_z =  camob.matrix_world[2][1]
    #cam_dir_up_x = camob.matrix_world[1][0]
    #cam_dir_up_y = camob.matrix_world[1][2]
    #cam_dir_up_z = -camob.matrix_world[1][1]
    cam_dir_x = -camob.matrix_world[0][2]
    cam_dir_y = -camob.matrix_world[2][2]
    cam_dir_z =  camob.matrix_world[1][2]
    cam_dir_up_x = camob.matrix_world[0][1]
    cam_dir_up_y = camob.matrix_world[2][1]
    cam_dir_up_z = -camob.matrix_world[1][1]
    cam_angle = degrees(camob.data.angle)
    if width < height:
        leng = 18.0/tan(camob.data.angle/2)
        cam_angle = 2*atan(18.0*width/height/leng)
        cam_angle = degrees(cam_angle)
        #cam_angle = cam_angle * width / height
    cam_active = 1
    cam_dof_on = 0
    cam_focus_dist = 1.0
    cam_aperture = 4.0
    cam_iso = 100.0
    cam_shutter = 50.0
    cam_bl_num = 5
    cam_bl_angle = 0.0
    cam_bl_round = 0
    cam_bl_radius = 1.0
    cam_bokeh_balance = 0
    cam_bokeh_size = 20
    cam_bokeh_flatten = 0
    cam_bokeh_vignette = 0
    
    if hasattr(sce, "noxData"):
        #width = sce.noxData.width
        #height = sce.noxData.height
        aa = sce.noxData.aa
    if hasattr(camob.data, "noxData"):
        noxcam = camob.data.noxData
        if noxcam.active:
            cam_active = 1
        else:
            cam_active = 0
        if noxcam.dof_on:
            cam_dof_on = 1
        else:
            cam_dof_on = 0
        if noxcam.ap_round_blades:
            cam_bl_round = 1
        else:
            cam_bl_round = 0
        cam_aperture = noxcam.aperture
        cam_iso = float(noxcam.iso)
        cam_shutter = noxcam.shutter
        cam_bl_num = noxcam.ap_num_blades
        cam_bl_angle = noxcam.ap_angle
        cam_bl_round = noxcam.ap_round_blades
        cam_bl_radius = noxcam.ap_blades_radius
        cam_bokeh_balance = noxcam.bokeh_ring_balance
        cam_bokeh_size = noxcam.bokeh_ring_size
        cam_bokeh_flatten = noxcam.bokeh_flatten
        cam_bokeh_vignette = noxcam.bokeh_vignette
    cam_focus_dist = camob.data.dof_distance
    if cam_focus_dist < 0.01:
        cam_focus_dist = 0.01
    if cam_focus_dist>1000000000.0:
        cam_focus_dist = 1000000000.0
            

    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_CAMERA)
    file.write(data) #2
    data=struct.pack("=Q", cam_size)
    file.write(data) #8
    data=struct.pack("=HH", NOX_BIN_CAM_ACTIVE, cam_active)
    file.write(data) #4
    data=struct.pack("=HI", NOX_BIN_CAM_WIDTH, width)
    file.write(data) #6
    data=struct.pack("=HI", NOX_BIN_CAM_HEIGHT, height)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_AA, aa)
    file.write(data) #4
    data=struct.pack("=Hfff", NOX_BIN_CAM_POS, camob.location[0], camob.location[2], -camob.location[1])
    file.write(data) #14
    data=struct.pack("=Hfff", NOX_BIN_CAM_DIR, cam_dir_x, cam_dir_y, cam_dir_z)
    file.write(data) #14
    data=struct.pack("=Hfff", NOX_BIN_CAM_UPDIR, cam_dir_up_x, cam_dir_up_y, cam_dir_up_z)
    file.write(data) #14
    data=struct.pack("=Hf", NOX_BIN_CAM_ANGLE, cam_angle)
    file.write(data) #6
   
    name_size = len(camob.name)+6
    data=struct.pack("=HI", NOX_BIN_CAM_NAME, name_size)
    file.write(data) #6
    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode(camob.name))
    file.write(data) #xxx
    data=struct.pack("=HH", NOX_BIN_CAM_DOF_ON, cam_dof_on)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_CAM_FOCUSDIST, cam_focus_dist)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_CAM_ISO, cam_iso)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_CAM_APERTURE, cam_aperture)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_CAM_SHUTTER, cam_shutter)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BLADESNUMBER, cam_bl_num)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_CAM_BLADESANGLE, cam_bl_angle)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BLADESROUND, cam_bl_round)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_CAM_BLADESRADIUS, cam_bl_radius)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BOKEH_RING_SIZE, cam_bokeh_size)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BOKEH_RING_BALANCE, cam_bokeh_balance)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BOKEH_FLATTEN, cam_bokeh_flatten)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_CAM_BOKEH_VIGNETTE, cam_bokeh_vignette)
    file.write(data) #6

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != cam_size):
        NOX_debug.addConsoleLog(logON, "Camera size incorrect: "+str(diff_pos)+" instead of "+str(cam_size)+".\nScene will be corrupted.")
	

#---------------------------------------------------------------------------------------------------

def add_cameras(sce, file):
    for ob in sce.objects:
        if (ob.type=="CAMERA"):
            add_camera(sce, file, ob)
    #for cam in bpy.data.cameras:
    #    add_camera(sce, file, cam)

#---------------------------------------------------------------------------------------------------

def add_default_material(sce, file):
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "adding simple material")

    mat_ID = 0
    num_layers = 1
    blend_layer = 0
    opacity = 100
    l_type = 2
    l_contr = 100
    l_diff0r = 0.8
    l_diff0g = 0.8
    l_diff0b = 0.8
    l_diff0a = 1.0
    l_diff90r = 0.9
    l_diff90g = 0.9
    l_diff90b = 0.9
    l_diff90a = 1.0
    l_rough = 1.0
    l_ior = 1.6
    l_transm_on = 0
    l_fake_on = 0
    l_disp_on = 0
    l_disp_val = 0.0
    l_abs_on = 0
    l_absr = 0.9
    l_absg = 0.7
    l_absb = 0.6
    l_absa = 1.0
    l_abs_dist = 0.1
    l_emissr = 0.9
    l_emissg = 0.7
    l_emissb = 0.6
    l_emissa = 1.0
    l_emisstemp = 6500
    l_emisspower = 100
    l_emissunit = 1

    name_size = 16

    mat_size = eval_material_default_size()
    layer_size = 152
    start_pos = file.tell()


    data=struct.pack("=HQIII", NOX_BIN_MATERIAL, mat_size, mat_ID, blend_layer, num_layers)
    file.write(data) #22
    data=struct.pack("=HI", NOX_BIN_MATNAME, name_size)
    file.write(data) #6
    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode("DefaultMat"))
    file.write(data) #10
    data=struct.pack("=Hf", NOX_BIN_MAT_OPACITY, opacity)
    file.write(data) #6
    data=struct.pack("=HQHI", NOX_BIN_MATLAYER, layer_size, l_type, l_contr)
    file.write(data) #16
    data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL0, l_diff0r, l_diff0g, l_diff0b, l_diff0a)
    file.write(data) #18
    data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL90, l_diff90r, l_diff90g, l_diff90b, l_diff90a)
    file.write(data) #18
    data=struct.pack("=Hf", NOX_BIN_MLAY_ROUGH, l_rough)
    file.write(data) #6
    data=struct.pack("=HIHf", NOX_BIN_MLAY_FRESNEL, 12, NOX_BIN_MLAY_FRES_IOR, l_ior)
    file.write(data) #12
    data=struct.pack("=HH", NOX_BIN_MLAY_TRANSM_ON, l_transm_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_FAKEGLASS_ON, l_fake_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_DISP_ON, l_disp_on)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_MLAY_DISP_VALUE, l_disp_val)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_MLAY_ABS_ON, l_abs_on)
    file.write(data) #4
    data=struct.pack("=Hffff", NOX_BIN_MLAY_ABS_COLOR, l_absr, l_absg, l_absb, l_absa)
    file.write(data) #18
    data=struct.pack("=Hf", NOX_BIN_MLAY_ABS_DIST, l_abs_dist)
    file.write(data) #6
    data=struct.pack("=Hffff", NOX_BIN_MLAY_EMISS_COLOR, l_emissr, l_emissg, l_emissb, l_emissa)
    file.write(data) #18
    data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_TEMP, l_emisstemp)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_EMISS_POWER, l_emisspower)
    file.write(data) #6
    data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_UNIT, l_emissunit)
    file.write(data) #6

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != mat_size):
        NOX_debug.addConsoleLog(logON, "Default material size incorrect: "+str(diff_pos)+" instead of "+str(mat_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def add_material(file, mnum):
    logON = NOX_debug.log_export_nox_scene
    mat = bpy.data.materials[mnum]
    NOX_debug.addConsoleLog(logON, "adding material "+str(mnum+1)+ " : "+mat.name)

    mat_ID = mnum+1
    num_layers = 1
    blend_layer = 0
    opacity = 100
    l_type = 2
    l_contr = 100
    l_diff0r = mat.diffuse_color[0]
    l_diff0g = mat.diffuse_color[1]
    l_diff0b = mat.diffuse_color[2]
    l_diff0a = 1.0
    l_diff90r = mat.specular_color[0]
    l_diff90g = mat.specular_color[0]
    l_diff90b = mat.specular_color[0]
    l_diff90a = 1.0
    l_rough = 1.0
    l_ior = 1.6
    l_transm_on = 0
    l_fake_on = 0
    l_disp_on = 0
    l_disp_val = 0.0
    l_abs_on = 0
    l_absr = 0.9
    l_absg = 0.7
    l_absb = 0.6
    l_absa = 1.0
    l_abs_dist = 0.1
    l_emissr = 0.9
    l_emissg = 0.7
    l_emissb = 0.6
    l_emissa = 1.0
    l_emisstemp = 6500
    l_emisspower = 100
    l_emissunit = 1

    mat_size = eval_material_size(mnum)
    name_size = 6+len(mat.name)
    layer_size = 152
    start_pos = file.tell()

    data=struct.pack("=HQIII", NOX_BIN_MATERIAL, mat_size, mat_ID, blend_layer, num_layers)
    file.write(data) #22
    data=struct.pack("=HI", NOX_BIN_MATNAME, name_size)
    file.write(data) #6
    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode(mat.name))
    file.write(data) #10
    data=struct.pack("=Hf", NOX_BIN_MAT_OPACITY, opacity)
    file.write(data) #6
    data=struct.pack("=HQHI", NOX_BIN_MATLAYER, layer_size, l_type, l_contr)
    file.write(data) #16
    data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL0, l_diff0r, l_diff0g, l_diff0b, l_diff0a)
    file.write(data) #18
    data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL90, l_diff90r, l_diff90g, l_diff90b, l_diff90a)
    file.write(data) #18
    data=struct.pack("=Hf", NOX_BIN_MLAY_ROUGH, l_rough)
    file.write(data) #6
    data=struct.pack("=HIHf", NOX_BIN_MLAY_FRESNEL, 12, NOX_BIN_MLAY_FRES_IOR, l_ior)
    file.write(data) #12
    data=struct.pack("=HH", NOX_BIN_MLAY_TRANSM_ON, l_transm_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_FAKEGLASS_ON, l_fake_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_DISP_ON, l_disp_on)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_MLAY_DISP_VALUE, l_disp_val)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_MLAY_ABS_ON, l_abs_on)
    file.write(data) #4
    data=struct.pack("=Hffff", NOX_BIN_MLAY_ABS_COLOR, l_absr, l_absg, l_absb, l_absa)
    file.write(data) #18
    data=struct.pack("=Hf", NOX_BIN_MLAY_ABS_DIST, l_abs_dist)
    file.write(data) #6
    data=struct.pack("=Hffff", NOX_BIN_MLAY_EMISS_COLOR, l_emissr, l_emissg, l_emissb, l_emissa)
    file.write(data) #18
    data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_TEMP, l_emisstemp)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_EMISS_POWER, l_emisspower)
    file.write(data) #6
    data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_UNIT, l_emissunit)
    file.write(data) #6

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != mat_size):
        NOX_debug.addConsoleLog(logON, "Material size incorrect: "+str(diff_pos)+" instead of "+str(mat_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def add_texture(file, mnum, lnum, mslot):
    global glb_scene
    logON = NOX_debug.log_export_nox_scene

    if mslot == TEX_SLOT_ENVMAP:
        world = glb_scene.world
        #world = bpy.context.scene.world
        if world:
            if hasattr(world, "noxDataEnv"):
                tslot = world.noxDataEnv.texture
    else:
        mat = bpy.data.materials[mnum]
        if not hasattr(mat, "noxData"):
            return 0

        if mslot == MAT_TEX_SLOT_OPACITY:
            tslot = mat.noxData.tex_opacity
        else:
            mlay = mat.noxData.layers[lnum]
            if mslot == LAYER_TEX_SLOT_COLOR0:
                tslot = mlay.tex_refl0
            if mslot == LAYER_TEX_SLOT_COLOR90:
                tslot = mlay.tex_refl90
            if mslot == LAYER_TEX_SLOT_ROUGHNESS:
                tslot = mlay.tex_rough
            if mslot == LAYER_TEX_SLOT_WEIGHT:
                tslot = mlay.tex_weight
            if mslot == LAYER_TEX_SLOT_LIGHT:
                tslot = mlay.tex_light
            if mslot == LAYER_TEX_SLOT_NORMAL:
                tslot = mlay.tex_normal

    fullfilename = ''
    if tslot.texID<bpy.data.textures.__len__() and tslot.texID>-1:
        tex = bpy.data.textures[tslot.texID]
        if hasattr(tex, "image") and tex.type=="IMAGE":
            if hasattr(tex.image, 'filepath') and hasattr(tex.image, 'source'):
                if tex.image.source=='FILE' and tex.image.filepath!='':
                    fullfilename = tex.image.filepath
    
    if fullfilename!='':
        absfullfilename = bpy.path.abspath(fullfilename)
        fnlen = 6 + len(absfullfilename)
    else:
        absfullfilename = ''
        fnlen = 0

    tex_size = eval_texture_size(mnum, lnum, mslot)
    tex_type = 0
    tex_slot = mslot
    if tslot.enabled:
        tex_enabled = 1
    else:
        tex_enabled = 0
    tex_brightness = tslot.brightness
    tex_contrast = tslot.contrast
    tex_saturation = tslot.saturation
    tex_hue = tslot.hue
    tex_red = tslot.red
    tex_green = tslot.green
    tex_blue = tslot.blue
    tex_gamma = tslot.gamma
    if tslot.invert:
        tex_invert = 1
    else:
        tex_invert = 0
    if tslot.bilinear:
        tex_filter = 1
    else:
        tex_filter = 0
    tex_precalc = 0
    post_size = 58

    start_pos = file.tell()
    data=struct.pack("=HQHH", NOX_BIN_MLAY_TEXTURE, tex_size, tex_type, tex_slot)
    file.write(data) #14
    if fnlen>0:
        data=struct.pack("=HI", NOX_BIN_MLAY_TEX_FILENAME, fnlen)
        file.write(data) #6
        strformat = str(fnlen-6)+"s"
        data=struct.pack(strformat, str.encode(absfullfilename))
        file.write(data) #XXX
    data=struct.pack("=HH", NOX_BIN_MLAY_TEX_ENABLED, tex_enabled)
    file.write(data) #4
    data=struct.pack("=HI", NOX_BIN_MLAY_TEX_POST, post_size)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_BRIGHTNESS, tex_brightness)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_CONTRAST, tex_contrast)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_SATURATION, tex_saturation)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_HUE, tex_hue)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_RED, tex_red)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_GREEN, tex_green)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_BLUE, tex_blue)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_MLAY_TEX_POST_GAMMA, tex_gamma)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_MLAY_TEX_POST_INVERT, tex_invert)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_TEX_FILTER, tex_filter)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_MLAY_TEX_PRECALCULATE, tex_precalc)
    file.write(data) #4
    
    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != tex_size):
        NOX_debug.addConsoleLog(logON, "Texture size incorrect: "+str(diff_pos)+" instead of "+str(tex_size)+".\nScene will be corrupted.")
    
#---------------------------------------------------------------------------------------------------

def is_nox_emitter(mnum):
    if mnum>=bpy.data.materials.__len__():
        return False

    mat = bpy.data.materials[mnum]

    if not hasattr(mat, "noxData"):
        return False
    if mat.noxData.layers.__len__() < 1:
        return False

    num_layers = mat.noxData.layers.__len__()
    for i in range(num_layers):
        mlay = mat.noxData.layers[i]
        if mlay.emission:
            return True

    return False
    
#---------------------------------------------------------------------------------------------------

def add_nox_material(file, mnum):
    logON = NOX_debug.log_export_nox_scene
    mat = bpy.data.materials[mnum]
    NOX_debug.addConsoleLog(logON, "adding nox material "+str(mnum+1)+ " : "+mat.name)
    
    if not hasattr(mat, "noxData"):
        add_material(file, mnum)
        return True

    if mat.noxData.layers.__len__() < 1:
        add_material(file, mnum)
        return True

    mat_ID = mnum+1
    num_layers = mat.noxData.layers.__len__()

    blen = mat.noxData.blend_layer.__len__()
    if blen==3:
        strblay = "{1}{2}".format(*mat.noxData.blend_layer)
        try:
            blend_layer = int(strblay)-1
        except ValueError:
            blend_layer = 0
    elif blen==2:
        strblay = "{1}".format(*mat.noxData.blend_layer)
        try:
            blend_layer = int(strblay)-1
        except ValueError:
            blend_layer = 0
    else:
        blend_layer = 0
    
    opacity = mat.noxData.opacity
    skyportal = 0
    if mat.noxData.skyportal:
        skyportal = 1

    mat_size = eval_material_nox_size(mnum)
    name_size = 6+len(mat.name)
    #layer_size = 152
    start_pos = file.tell()

    data=struct.pack("=HQIII", NOX_BIN_MATERIAL, mat_size, mat_ID, blend_layer, num_layers)
    file.write(data) #22
    data=struct.pack("=HI", NOX_BIN_MATNAME, name_size)
    file.write(data) #6
    strformat = str(name_size-6)+"s"
    data=struct.pack(strformat, str.encode(mat.name))
    file.write(data) #XXX
    data=struct.pack("=Hf", NOX_BIN_MAT_OPACITY, opacity)
    file.write(data) #6
    
    data=struct.pack("=HH", NOX_BIN_MAT_SKYPORTAL, skyportal)
    file.write(data) #4

    add_texture(file, mnum, 0, MAT_TEX_SLOT_OPACITY)
    
    for i in range(num_layers):
        layer_size = eval_material_layer_nox_size(mnum, i)
        layer_start_pos = file.tell()
        mlay = mat.noxData.layers[i]
        if mlay.emission:
            l_type = LAYER_TYPE_EMITTER
        else:
            l_type = LAYER_TYPE_SHADE
        l_contr = int(mlay.weight)
        l_diff0r = mlay.refl0[0]
        l_diff0g = mlay.refl0[1]
        l_diff0b = mlay.refl0[2]
        l_diff0a = 1.0
        l_diff90r = mlay.refl90[0]
        l_diff90g = mlay.refl90[1]
        l_diff90b = mlay.refl90[2]
        l_diff90a = 1.0
        l_rough = mlay.rough/100.0
        l_ior = mlay.ior
        if mlay.transmission:
            l_transm_on = 1
        else:
            l_transm_on = 0
        if mlay.fake_glass:
            l_fake_on = 1
        else:
            l_fake_on = 0
        if mlay.dispersion:
            l_disp_on = 1
        else:
            l_disp_on = 0
        l_disp_val = mlay.dispersion_size
        if mlay.absorption:
            l_abs_on = 1
        else:
            l_abs_on = 0
        l_absr = mlay.absorption_color[0]
        l_absg = mlay.absorption_color[1]
        l_absb = mlay.absorption_color[2]
        l_absa = 1.0
        l_abs_dist = mlay.absorption_distance
        l_emissr = mlay.emission_color[0]
        l_emissg = mlay.emission_color[1]
        l_emissb = mlay.emission_color[2]
        l_emissa = 1.0
        l_emisstemp = mlay.emission_temperature
        l_emisspower = mlay.emission_power
        l_emissunit = EMISSION_UNIT_WATT
        if mlay.emission_unit == "Wm":
            l_emissunit = EMISSION_UNIT_WATT_PER_SQR_METER
        if mlay.emission_unit == "lm":
            l_emissunit = EMISSION_UNIT_LUMEN
        if mlay.emission_unit == "lx":
            l_emissunit = EMISSION_UNIT_LUX
        
        data=struct.pack("=HQHI", NOX_BIN_MATLAYER, layer_size, l_type, l_contr)
        file.write(data) #16
        data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL0, l_diff0r, l_diff0g, l_diff0b, l_diff0a)
        file.write(data) #18
        data=struct.pack("=Hffff", NOX_BIN_MLAY_REFL90, l_diff90r, l_diff90g, l_diff90b, l_diff90a)
        file.write(data) #18
        data=struct.pack("=Hf", NOX_BIN_MLAY_ROUGH, l_rough)
        file.write(data) #6
        data=struct.pack("=HIHf", NOX_BIN_MLAY_FRESNEL, 12, NOX_BIN_MLAY_FRES_IOR, l_ior)
        file.write(data) #12
        data=struct.pack("=HH", NOX_BIN_MLAY_TRANSM_ON, l_transm_on)
        file.write(data) #4
        data=struct.pack("=HH", NOX_BIN_MLAY_FAKEGLASS_ON, l_fake_on)
        file.write(data) #4
        data=struct.pack("=HH", NOX_BIN_MLAY_DISP_ON, l_disp_on)
        file.write(data) #4
        data=struct.pack("=Hf", NOX_BIN_MLAY_DISP_VALUE, l_disp_val)
        file.write(data) #6
        data=struct.pack("=HH", NOX_BIN_MLAY_ABS_ON, l_abs_on)
        file.write(data) #4
        data=struct.pack("=Hffff", NOX_BIN_MLAY_ABS_COLOR, l_absr, l_absg, l_absb, l_absa)
        file.write(data) #18
        data=struct.pack("=Hf", NOX_BIN_MLAY_ABS_DIST, l_abs_dist)
        file.write(data) #6
        data=struct.pack("=Hffff", NOX_BIN_MLAY_EMISS_COLOR, l_emissr, l_emissg, l_emissb, l_emissa)
        file.write(data) #18
        data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_TEMP, l_emisstemp)
        file.write(data) #6
        data=struct.pack("=Hf", NOX_BIN_MLAY_EMISS_POWER, l_emisspower)
        file.write(data) #6
        data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_UNIT, l_emissunit)
        file.write(data) #6

        if mlay.ies_file.__len__()>0:
            absfullfilename = bpy.path.abspath(mlay.ies_file)
            fnlen = len(absfullfilename)+6
            if fnlen>0:
                data=struct.pack("=HI", NOX_BIN_MLAY_EMISS_IES, fnlen)
                file.write(data) #6
                strformat = str(fnlen-6)+"s"
                data=struct.pack(strformat, str.encode(absfullfilename))
                file.write(data) #XXX

        add_texture(file, mnum, i, LAYER_TEX_SLOT_COLOR0)
        add_texture(file, mnum, i, LAYER_TEX_SLOT_COLOR90)
        add_texture(file, mnum, i, LAYER_TEX_SLOT_ROUGHNESS)
        add_texture(file, mnum, i, LAYER_TEX_SLOT_WEIGHT)
        add_texture(file, mnum, i, LAYER_TEX_SLOT_LIGHT)
        add_texture(file, mnum, i, LAYER_TEX_SLOT_NORMAL)
        
    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != mat_size):
        NOX_debug.addConsoleLog(logON, "Material size incorrect: "+str(diff_pos)+" instead of "+str(mat_size)+".\nScene will be corrupted.")
	
#---------------------------------------------------------------------------------------------------

def add_materials(sce, file):
    global glb_mat_override
    global glb_emitter_exists
    logON = NOX_debug.log_export_nox_scene
    NOX_debug.addConsoleLog(logON, "adding materials")

    num_mats = bpy.data.materials.__len__()+1

    mats_size = eval_materials_size(sce)
    start_pos = file.tell()

    data=struct.pack("=H", NOX_BIN_MATERIALS)
    file.write(data)
    data=struct.pack("=Q", mats_size)
    file.write(data)
    data=struct.pack("=I", num_mats)
    file.write(data)

    if glb_mat_override>-1:
        gmat = bpy.data.materials[glb_mat_override]
        if hasattr(gmat, "noxData"):
            add_nox_material(file, glb_mat_override)
            if is_nox_emitter(glb_mat_override):
                glb_emitter_exists = True
        else:
            add_material(file, glb_mat_override)
    else:
        add_default_material(sce, file)

    for i in range(num_mats-1):
        mat = bpy.data.materials[i]
        use_this = True
        if glb_mat_override > -1:
            if not is_nox_emitter(i):
                use_this = False
        if use_this:
            if hasattr(mat, "noxData"):
                add_nox_material(file, i)
            else:
                add_material(file, i)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != mats_size):
        NOX_debug.addConsoleLog(logON, "Materials size incorrect: "+str(diff_pos)+" instead of "+str(mats_size)+".\nScene will be corrupted.")

#---------------------------------------------------------------------------------------------------

def add_blend_layers(sce, file):
    logON = NOX_debug.log_export_nox_scene

    if not hasattr(sce, "noxData"):
        return True

    NOX_debug.addConsoleLog(logON, "adding blend layers")
        
    bl_size = eval_blend_layers_size(sce)
    start_pos = file.tell()
    
    data=struct.pack("=H", NOX_BIN_BLEND_LAYERS)
    file.write(data)
    data=struct.pack("=Q", bl_size)
    file.write(data)
    
    for i in range(16):
        bname = ""
        if i==0:
            bname = sce.noxData.blend_layers.b1
        if i==1:
            bname = sce.noxData.blend_layers.b2
        if i==2:
            bname = sce.noxData.blend_layers.b3
        if i==3:
            bname = sce.noxData.blend_layers.b4
        if i==4:
            bname = sce.noxData.blend_layers.b5
        if i==5:
            bname = sce.noxData.blend_layers.b6
        if i==6:
            bname = sce.noxData.blend_layers.b7
        if i==7:
            bname = sce.noxData.blend_layers.b8
        if i==8:
            bname = sce.noxData.blend_layers.b9
        if i==9:
            bname = sce.noxData.blend_layers.b10
        if i==10:
            bname = sce.noxData.blend_layers.b11
        if i==11:
            bname = sce.noxData.blend_layers.b12
        if i==12:
            bname = sce.noxData.blend_layers.b13
        if i==13:
            bname = sce.noxData.blend_layers.b14
        if i==14:
            bname = sce.noxData.blend_layers.b15
        if i==15:
            bname = sce.noxData.blend_layers.b16

        strsize = len(bname)+6
        bnsize = strsize + 14
        data=struct.pack("=H", NOX_BIN_BLEND_LAYER)
        file.write(data)
        data=struct.pack("=Q", bnsize)
        file.write(data)
        data=struct.pack("=I", i)
        file.write(data)
        data=struct.pack("=H", NOX_BIN_BLEND_LAYER_NAME)
        file.write(data)
        data=struct.pack("=I", strsize)
        file.write(data)
        strformat = str(strsize-6)+"s"
        data=struct.pack(strformat, str.encode(bname))
        file.write(data) #XXX

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != bl_size):
        NOX_debug.addConsoleLog(logON, "Blend layers size incorrect: "+str(diff_pos)+" instead of "+str(bl_size)+".\nScene will be corrupted.")

    return True

#---------------------------------------------------------------------------------------------------

def add_sunsky(sce, file):
    global glb_emitter_exists
    logON = NOX_debug.log_export_nox_scene
    world = sce.world
    if not world:
        return True
    if not hasattr(world, "noxData"):
        return True

    NOX_debug.addConsoleLog(logON, "adding sunsky")
    sunsky_size = eval_sunsky_size(sce)
    start_pos = file.tell()

    if world.noxData.active  or  world.noxData.sun_on:
        glb_emitter_exists = True
    
    sunsky_on = 1 if world.noxData.active else 0
    month = world.noxData.month
    day = world.noxData.day
    hour = world.noxData.hour
    minute = world.noxData.minute
    gmt = world.noxData.gmt
    longitude = -world.noxData.longitude*math.pi/180.0
    latitude = -world.noxData.latitude*math.pi/180.0
    turbidity = world.noxData.turbidity/10.0
    aerosol = 0
    sun_size = world.noxData.sun_size
    sun_on = 1 if world.noxData.sun_on else 0
    sun_other_layer = 1 if world.noxData.sun_layer2 else 0
    angle = world.noxData.north_angle * math.pi/180.0
    dir_n_x = math.sin(angle)
    dir_n_y = 0
    dir_n_z = -math.cos(angle)
    dir_e_x = math.cos(angle)
    dir_e_y = 0
    dir_e_z = math.sin(angle)
    
    
    data=struct.pack("=HQ", NOX_BIN_SUNSKY, sunsky_size)
    file.write(data) #10
    data=struct.pack("=HH", NOX_BIN_SS_ON, sunsky_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_SS_MONTH, month)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_SS_DAY, day)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_SS_HOUR, hour)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_SS_MINUTE, minute)
    file.write(data) #4
    data=struct.pack("=Hh", NOX_BIN_SS_GMT, gmt)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_SS_LONGITUDE, longitude)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_SS_LATITUDE, latitude)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_SS_TURBIDITY, turbidity)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_SS_AEROSOL, aerosol)
    file.write(data) #6
    data=struct.pack("=Hf", NOX_BIN_SS_SUN_SIZE, sun_size)
    file.write(data) #6
    data=struct.pack("=HH", NOX_BIN_SS_SUN_ON, sun_on)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_SS_SUN_OTHER_LAYER, sun_other_layer)
    file.write(data) #4
    data=struct.pack("=Hffffff", NOX_BIN_SS_DIRECTIONS, dir_n_x, dir_n_y, dir_n_z, dir_e_x, dir_e_y, dir_e_z)
    file.write(data) #26


    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != sunsky_size):
        NOX_debug.addConsoleLog(logON, "Sunsky size incorrect: "+str(diff_pos)+" instead of "+str(sunsky_size)+".\nScene will be corrupted.")

    return True
    
#---------------------------------------------------------------------------------------------------

def add_envmap(sce, file):
    global glb_emitter_exists
    logON = NOX_debug.log_export_nox_scene
    world = sce.world
    if not world:
        return True
    if not hasattr(world, "noxDataEnv"):
        return True

    if world.noxDataEnv.texture.enabled  and  world.noxDataEnv.texture.texID>-1:
        glb_emitter_exists = True

    NOX_debug.addConsoleLog(logON, "adding environment map")
    env_size = eval_envmap_size(sce)
    start_pos = file.tell()

    env_power = world.noxDataEnv.power
    env_blend = world.noxDataEnv.blendlayer
    env_azimuth = world.noxDataEnv.azimuth
    env_enabled = world.noxDataEnv.texture.enabled
    dir_n_x = 0.0
    dir_n_y = 0.0
    dir_n_z = -1.0
    dir_e_x = 1.0
    dir_e_y = 0.0
    dir_e_z = 0.0
    
    data=struct.pack("=HQ", NOX_BIN_ENVIRONMENT_MAP, env_size)
    file.write(data) #10
    data=struct.pack("=HH", NOX_BIN_ENV_ENABLED, env_enabled)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_ENV_POWER_EV, env_power)
    file.write(data) #4
    data=struct.pack("=Hf", NOX_BIN_ENV_AZIMUTH, env_azimuth)
    file.write(data) #4
    data=struct.pack("=HH", NOX_BIN_ENV_BLEND_LAYER, env_blend)
    file.write(data) #4
    data=struct.pack("=Hffffff", NOX_BIN_ENV_DIRECTIONS, dir_n_x, dir_n_y, dir_n_z, dir_e_x, dir_e_y, dir_e_z)
    file.write(data) #26
    
    add_texture(file, 0, 0, TEX_SLOT_ENVMAP)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != env_size):
        NOX_debug.addConsoleLog(logON, "Sunsky size incorrect: "+str(diff_pos)+" instead of "+str(env_size)+".\nScene will be corrupted.")

    return True
    
#---------------------------------------------------------------------------------------------------

def add_preload_stuff(sce, file):
    if hasattr(sce, "noxData"):
        if len(sce.noxData.preload_post_file)>0  and  sce.noxData.preload_post_on:
            fullfilename = sce.noxData.preload_post_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            if fnlen>0:
                data=struct.pack("=HI", NOX_BIN_PRELOAD_POST, fnlen)
                file.write(data) #6
                strformat = str(fnlen-6)+"s"
                data=struct.pack(strformat, str.encode(absfullfilename))
                file.write(data) #XXX
        if len(sce.noxData.preload_blend_file)>0  and  sce.noxData.preload_blend_on:
            fullfilename = sce.noxData.preload_blend_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            if fnlen>0:
                data=struct.pack("=HI", NOX_BIN_PRELOAD_BLEND, fnlen)
                file.write(data) #6
                strformat = str(fnlen-6)+"s"
                data=struct.pack(strformat, str.encode(absfullfilename))
                file.write(data) #XXX
        if len(sce.noxData.preload_final_file)>0  and  sce.noxData.preload_final_on:
            fullfilename = sce.noxData.preload_final_file
            absfullfilename = bpy.path.abspath(fullfilename)
            fnlen = 6 + len(absfullfilename)
            if fnlen>0:
                data=struct.pack("=HI", NOX_BIN_PRELOAD_FINAL, fnlen)
                file.write(data) #6
                strformat = str(fnlen-6)+"s"
                data=struct.pack(strformat, str.encode(absfullfilename))
                file.write(data) #XXX
    
#---------------------------------------------------------------------------------------------------

def add_scene(sce, file):
    logON = NOX_debug.log_export_nox_scene
    global glb_scene_size

    start_pos = file.tell()
    data=struct.pack("=H", NOX_BIN_SCENE)
    file.write(data)
    data=struct.pack("=Q", glb_scene_size)
    file.write(data)

    add_preload_stuff(sce, file)
    
    add_geometry(sce, file)

    add_cameras(sce, file)

    add_materials(sce, file)
    
    add_blend_layers(sce, file)
    
    add_sunsky(sce, file)

    add_envmap(sce, file)

    end_pos = file.tell()
    diff_pos = end_pos - start_pos
    if (diff_pos != glb_scene_size):
        NOX_debug.addConsoleLog(logON, "Scene size incorrect: "+str(diff_pos)+" instead of "+str(glb_scene_size)+".\nScene will be corrupted.")


#---------------------------------------------------------------------------------------------------

def write_NOX_scene(filename, context):
    logON = NOX_debug.log_export_nox_scene
    global glb_scene_size
    global glb_geom_size
    global glb_num_meshes
    global glb_num_tris
    global glb_mat_override
    global glb_scene
    global glb_emitter_exists
    glb_scene_size = 0
    glb_geom_size = 0
    glb_num_meshes = 0
    glb_num_tris = 0

    sce = context.scene
    glb_scene = sce
    if not sce:
        bpy.ops.error.message('INVOKE_DEFAULT', message="Export error. Scene is null. Please try again.")
        NOX_debug.addConsoleLog(logON, "-------------------------------------------")
        NOX_debug.addConsoleLog(logON, "ERROR ---- Exporting error: scene is null ")
        NOX_debug.addConsoleLog(logON, "Please try again ....")
        NOX_debug.addConsoleLog(logON, "-------------------------------------------")
        return

    #find override material
    glb_mat_override = -1
    renderlayer = sce.render.layers[0]
    omat = renderlayer.material_override
    if omat != None  and  sce.noxData.override_on:
        num_mats = bpy.data.materials.__len__()+1
        for i in range(num_mats-1):
            tmat = bpy.data.materials[i]
            if tmat.name == omat.name:
                glb_mat_override = i
        NOX_debug.addConsoleLog(logON, "material override: "+omat.name+" - id: "+str(glb_mat_override))
        
    NOX_debug.addConsoleLog(logON, "\nExporting NOX scene: " + filename)
    NOX_debug.addConsoleLog(logON, "Evaluating scene size...")
        
    glb_emitter_exists = False

    scene_size = eval_scene_size(sce)
    file_size = 0
    file_size = scene_size+16

    NOX_debug.addConsoleLog(logON, "Creating file...")
    out = open(filename, "wb")

    # write header, size, version
    #data=struct.pack("=BBBB", 78, 79, 88, 32)
    data=struct.pack("=I", NOX_BIN_HEADER)
    out.write(data)
    data=struct.pack("=Q", file_size)
    out.write(data)
    data=struct.pack("=BBBB", 0, nox_version_major, nox_version_minor, nox_version_build)
    out.write(data)

    # add scene and its internals
    add_scene(sce, out)

    if not glb_emitter_exists:
        bpy.ops.error.message('INVOKE_DEFAULT', message="Warning. No emitter on scene.")
        NOX_debug.addConsoleLog(logON, "-------------------------------------------")
        NOX_debug.addConsoleLog(logON, "WARNING - no emitter on scene")
        NOX_debug.addConsoleLog(logON, "-------------------------------------------")

    NOX_debug.addConsoleLog(logON, "Closing file...")
    out.close()

