import bpy
from mathutils import Color
from xml.dom import minidom
import xml
import struct
#from io_utils import ExportHelper
from bpy_extras.io_utils import ExportHelper
from bpy.props import StringProperty, BoolProperty, EnumProperty

nox_xml_error_node = ""

#----------------------------------------------------------------

def export_material(filename, mnum):
    global nox_xml_error_node
    if mnum<0 or mnum >= bpy.data.materials.__len__():
        return False
    
    mat = bpy.data.materials[mnum]
    if not hasattr(mat, "noxData"):
        return False

    doc = minidom.Document()

    # <material>    
    root_el = doc.createElement("material")
    root_el.setAttribute("id", "00000000")
    doc.appendChild(root_el)

    # <name>    
    matname = mat.name
    node_name = doc.createElement("name")
    root_el.appendChild(node_name)
    tempnode = doc.createTextNode(matname)
    node_name.appendChild(tempnode)
    
    # <opacity>
    opac = mat.noxData.opacity
    node_opacity = doc.createElement("opacity")
    root_el.appendChild(node_opacity)
    tempnode = doc.createTextNode(str(opac))
    node_opacity.appendChild(tempnode)
    
    # <skyportal>
    skp_act = "no"
    if mat.noxData.skyportal:
        skp_act = "yes"
    node_skyportal = doc.createElement("skyportal")
    root_el.appendChild(node_skyportal)
    tempnode = doc.createTextNode(skp_act)
    node_skyportal.appendChild(tempnode)

    # <blendlayer>
    blay = mat.noxData.blend_layer
    node_blendlayer = doc.createElement("blendlayer")
    root_el.appendChild(node_blendlayer)
    if blay.__len__()==3:
        strblay = "{1}{2}".format(*blay)
    elif blay.__len__()==2:
        strblay = "{1}".format(*blay)
    else:
        strblay = "1"
    tempnode = doc.createTextNode(strblay)
    node_blendlayer.appendChild(tempnode)

    # ----- L A Y E R S ---------
    
    numlayers = mat.noxData.layers.__len__()
    if numlayers == 0:
        mat.noxData.layers.add()
    
    numlayers = mat.noxData.layers.__len__()
    for i in range(numlayers):
        layer = mat.noxData.layers[i]
        node_layer = doc.createElement("layer")
        if layer.emission==True:
            node_layer.setAttribute("type", "emission")
        else:
            node_layer.setAttribute("type", "shade")
        node_layer.setAttribute("contribution", str(int(layer.weight)))
        root_el.appendChild(node_layer)
        
        # <emission>
        emission_col = packHexFromColor(layer.emission_color)
        node_emission_col = doc.createElement("emission")
        node_layer.appendChild(node_emission_col)
        tempnode = doc.createTextNode(emission_col)
        node_emission_col.appendChild(tempnode)

        # <power>
        hexpower = packHexFromFloat(layer.emission_power)
        node_power = doc.createElement("power")
        node_layer.appendChild(node_power)
        tempnode = doc.createTextNode(hexpower)
        node_power.appendChild(tempnode)
        
        # <unit>
        e_unit = "w"
        if layer.emission_unit=="Wm":
            e_unit="w/m2"
        if layer.emission_unit=="lm":
            e_unit="lm"
        if layer.emission_unit=="lx":
            e_unit="lx"
        node_unit = doc.createElement("unit")
        node_layer.appendChild(node_unit)
        tempnode = doc.createTextNode(e_unit)
        node_unit.appendChild(tempnode)
            
        # <temperature>
        node_temperature = doc.createElement("temperature")
        node_layer.appendChild(node_temperature)
        tempnode = doc.createTextNode(str(layer.emission_temperature))
        node_temperature.appendChild(tempnode)
        
        # <ies>
        if layer.ies_file.__len__()>0:
            node_ies = doc.createElement("ies")
            node_layer.appendChild(node_ies)
            tempnode = doc.createTextNode(layer.ies_file)
            node_ies.appendChild(tempnode)

        # <diffuse>
        diff0_col = packHexFromColor(layer.refl0)
        node_diffuse = doc.createElement("diffuse")
        node_layer.appendChild(node_diffuse)
        tempnode = doc.createTextNode(diff0_col)
        node_diffuse.appendChild(tempnode)

        # <diffuse90>
        diff90_col = packHexFromColor(layer.refl90)
        node_diffuse90 = doc.createElement("diffuse90")
        node_layer.appendChild(node_diffuse90)
        tempnode = doc.createTextNode(diff90_col)
        node_diffuse90.appendChild(tempnode)
    
        # <roughness>
        node_roughness = doc.createElement("roughness")
        node_layer.appendChild(node_roughness)
        tempnode = doc.createTextNode(str(layer.rough))
        node_roughness.appendChild(tempnode)
    
        # <fresnel>
        node_fresnel = doc.createElement("fresnel")
        node_layer.appendChild(node_fresnel)

        # <ior>
        node_ior = doc.createElement("ior")
        node_fresnel.appendChild(node_ior)
        tempnode = doc.createTextNode(str(layer.ior))
        node_ior.appendChild(tempnode)
        
        # <transmission>
        tr_act = "no"
        if layer.transmission:
            tr_act = "yes"
        node_transmission = doc.createElement("transmission")
        node_transmission.setAttribute("active", tr_act)
        node_layer.appendChild(node_transmission)
        
        # <fakeglass>
        fg_act = "no"
        if layer.fake_glass:
            fg_act = "yes"
        node_fakeglass = doc.createElement("fakeglass")
        node_transmission.appendChild(node_fakeglass)
        tempnode = doc.createTextNode(fg_act)
        node_fakeglass.appendChild(tempnode)

        # <absorption>
        abs_act = "no"
        if layer.absorption:
            abs_act = "yes"
        node_absorption = doc.createElement("absorption")
        node_absorption.setAttribute("active", abs_act)
        node_transmission.appendChild(node_absorption)
        
        # <abscolor>
        abs_col = packHexFromColor(layer.absorption_color)
        node_abscol = doc.createElement("abscolor")
        node_absorption.appendChild(node_abscol)
        tempnode = doc.createTextNode(abs_col)
        node_abscol.appendChild(tempnode)
        
        # <distance>
        node_absdist = doc.createElement("distance")
        node_absorption.appendChild(node_absdist)
        tempnode = doc.createTextNode(str(layer.absorption_distance))
        node_absdist.appendChild(tempnode)

        # <dispersion_on>
        dp_act = "no"
        if layer.dispersion:
            dp_act = "yes"
        node_dispersion = doc.createElement("dispersion_on")
        node_transmission.appendChild(node_dispersion)
        tempnode = doc.createTextNode(dp_act)
        node_dispersion.appendChild(tempnode)
        
        # <dispersion_val>
        node_dispersion_size = doc.createElement("dispersion_val")
        node_transmission.appendChild(node_dispersion_size)
        tempnode = doc.createTextNode(str(layer.dispersion_size))
        node_dispersion_size.appendChild(tempnode)
        
        texOK = insertTexture(doc, node_layer, "refl0", mat, i)
        if not texOK:
            return False
        texOK = insertTexture(doc, node_layer, "refl90", mat, i)
        if not texOK:
            return False
        texOK = insertTexture(doc, node_layer, "rough", mat, i)
        if not texOK:
            return False
        texOK = insertTexture(doc, node_layer, "light", mat, i)
        if not texOK:
            return False
        texOK = insertTexture(doc, node_layer, "weight", mat, i)
        if not texOK:
            return False
        texOK = insertTexture(doc, node_layer, "normal", mat, i)
        if not texOK:
            return False
        
    texOK = insertTexture(doc, root_el, "opacity", mat, 0)
    if not texOK:
        return False

    insertPreview(doc, root_el, mat)
   
    # hack writexml
    oldwritexml = xml.dom.minidom.Element.writexml
    xml.dom.minidom.Element.writexml = fixed_writexml

    file = open(filename, "w")
    doc.writexml(file, "", "  ", "\n") # self.doc.writexml( self.filename )
    
    xml.dom.minidom.Element.writexml = oldwritexml
    
    return True

#----------------------------------------------------------------

def insertPreview(doc, node_mat, mat):
    if mat.noxData.preview_id<0 or mat.noxData.preview_id>=bpy.data.images.__len__():
        return True

    img = bpy.data.images[mat.noxData.preview_id]
    if img.size[0]!=192 or img.size[1]!=192:
        return True
    if not hasattr(img, "pixels"):
        return True
    if img.pixels.__len__() != 192*192*4:
        return True

    allpixels = ""
    ii = 0
    for p in img.pixels:
        ii+=1
        if ii%4!=0:
            f = 255.0*p
            b = int(f)
            h = hex(b)
            if b>15:
                allpixels += h[2]+h[3]
            else:
                allpixels += "0"+h[2]
    
    # <preview>
    node_preview = doc.createElement("preview")
    node_preview.setAttribute("width", "192")
    node_preview.setAttribute("height", "192")
    node_mat.appendChild(node_preview)
    
    
    for i in range(192):
        nstr = "l"+str(i)
        sstart = i*192*6
        sstop = (i+1)*192*6
        xmlline = allpixels[sstart:sstop]
        
        # <lXXX>    
        node_lxx = doc.createElement(nstr)
        node_preview.appendChild(node_lxx)
        tempnode = doc.createTextNode(xmlline)
        node_lxx.appendChild(tempnode)

    return True

#----------------------------------------------------------------

def insertTexture(doc, node_parent, type, mat, clay):
    node_name = ''
    if type=="opacity":
        tslot = mat.noxData.tex_opacity
        node_name = 'tex_opacity'
    else:
        if clay>=mat.noxData.layers.__len__():
            return False
        layer = mat.noxData.layers[clay]
        if type=="refl0":
            tslot = layer.tex_refl0
            node_name = 'tex_refl0'
        if type=="refl90":
            tslot = layer.tex_refl90
            node_name = 'tex_refl90'
        if type=="rough":
            tslot = layer.tex_rough
            node_name = 'tex_rough'
        if type=="light":
            tslot = layer.tex_light
            node_name = 'tex_light'
        if type=="weight":
            tslot = layer.tex_weight
            node_name = 'tex_weight'
        if type=="normal":
            tslot = layer.tex_normal
            node_name = 'tex_normal'

    if node_name == '':
        return False

    valid_tex = False
    fullfilename = ''
    if tslot.texID<bpy.data.textures.__len__() and tslot.texID>-1:
        tex = bpy.data.textures[tslot.texID]
        if hasattr(tex, "image") and tex.type=="IMAGE":
            if hasattr(tex.image, 'filepath') and hasattr(tex.image, 'source'):
                if tex.image.source=='FILE' and tex.image.filepath!='':
                    valid_tex = True
                    fullfilename = tex.image.filepath

    # <tex_XXXX type="file" enabled="yes/no">
    node_tex = doc.createElement(node_name)
    node_tex.setAttribute("type", "file")
    if tslot.enabled==True:
        node_tex.setAttribute("enabled", "yes")
    else:
        node_tex.setAttribute("enabled", "no")
    node_parent.appendChild(node_tex)
  
    # <filename>
    if fullfilename!='':
        absfullfilename = bpy.path.abspath(fullfilename)
        node_filename = doc.createElement("filename")
        node_tex.appendChild(node_filename)
        tempnode = doc.createTextNode(absfullfilename)
        node_filename.appendChild(tempnode)
        
    # <filter>
    node_filter = doc.createElement("filter")
    node_tex.appendChild(node_filter)
    if tslot.bilinear==True:
        tempnode = doc.createTextNode("1")
    else:
        tempnode = doc.createTextNode("0")
    node_filter.appendChild(tempnode)

    # <precalc>
    node_precalc = doc.createElement("precalc")
    node_tex.appendChild(node_precalc)
    tempnode = doc.createTextNode("0")
    node_precalc.appendChild(tempnode)

    # <post>
    node_post = doc.createElement("post")
    node_tex.appendChild(node_post)
    
    # <brightness>
    node_brightness = doc.createElement("brightness")
    node_post.appendChild(node_brightness)
    tempnode = doc.createTextNode(str(tslot.brightness))
    node_brightness.appendChild(tempnode)
    
    # <contrast>
    node_contrast = doc.createElement("contrast")
    node_post.appendChild(node_contrast)
    tempnode = doc.createTextNode(str(tslot.contrast))
    node_contrast.appendChild(tempnode)

    # <saturation>
    node_saturation = doc.createElement("saturation")
    node_post.appendChild(node_saturation)
    tempnode = doc.createTextNode(str(tslot.saturation))
    node_saturation.appendChild(tempnode)

    # <hue>
    node_hue = doc.createElement("hue")
    node_post.appendChild(node_hue)
    tempnode = doc.createTextNode(str(tslot.hue))
    node_hue.appendChild(tempnode)
    
    # <red>
    node_red = doc.createElement("red")
    node_post.appendChild(node_red)
    tempnode = doc.createTextNode(str(tslot.red))
    node_red.appendChild(tempnode)

    # <green>
    node_green = doc.createElement("green")
    node_post.appendChild(node_green)
    tempnode = doc.createTextNode(str(tslot.green))
    node_green.appendChild(tempnode)

    # <blue>
    node_blue = doc.createElement("blue")
    node_post.appendChild(node_blue)
    tempnode = doc.createTextNode(str(tslot.blue))
    node_blue.appendChild(tempnode)

    # <gamma>
    node_gamma = doc.createElement("gamma")
    node_post.appendChild(node_gamma)
    tempnode = doc.createTextNode(str(tslot.gamma))
    node_gamma.appendChild(tempnode)

    # <invert>
    node_invert = doc.createElement("invert")
    node_post.appendChild(node_invert)
    if tslot.invert:
        tempnode = doc.createTextNode("yes")
    else:
        tempnode = doc.createTextNode("no")
    node_invert.appendChild(tempnode)
    
    return True

#----------------------------------------------------------------

def packHexFromFloat(val):
    aastr = struct.pack("=f", val)
    bt = struct.unpack("=BBBB", aastr)
    if bt[0]>15:
        h1 = "{3}{2}".format(*hex(bt[0]))
    else:
        h1 = "{2}0".format(*hex(bt[0]))
    if bt[1]>15:
        h2 = "{3}{2}".format(*hex(bt[1]))
    else:
        h2 = "{2}0".format(*hex(bt[1]))
    if bt[2]>15:
        h3 = "{3}{2}".format(*hex(bt[2]))
    else:
        h3 = "{2}0".format(*hex(bt[2]))
    if bt[3]>15:
        h4 = "{3}{2}".format(*hex(bt[3]))
    else:
        h4 = "{2}0".format(*hex(bt[3]))
    res = h1+h2+h3+h4
    return res

#----------------------------------------------------------------

def packHexFromColor(col):
    fr = min(255.0, max(0.0, col.r*255.0))
    fg = min(255.0, max(0.0, col.g*255.0))
    fb = min(255.0, max(0.0, col.b*255.0))
    ir = int(fr)
    ig = int(fg)
    ib = int(fb)
    if ir>15:
        hr = "{2}{3}".format(*hex(ir))
    else:
        hr = "0{2}".format(*hex(ir))
    if ig>15:
        hg = "{2}{3}".format(*hex(ig))
    else:
        hg = "0{2}".format(*hex(ig))
    if ib>15:
        hb = "{2}{3}".format(*hex(ib))
    else:
        hb = "0{2}".format(*hex(ib))
    res = hr+hg+hb
    return res

#----------------------------------------------------------------

class NOXMaterialExporter(bpy.types.Operator, ExportHelper):
    bl_idname = "wm.buttonexportnoxmaterial"
    bl_label = "Export NOX Material"
    filename_ext = ".nxm"
    filter_glob = StringProperty(default="*.nxm", options={'HIDDEN'})

    #addTex = BoolProperty(name="Copy textures", 
    #    description="Copy used textures - not implemented", default=False)

    
    def invoke(self, context, event):
        wm = context.window_manager
        wm.fileselect_add(self)
        return {"RUNNING_MODAL"}

    def execute(self, context):
        matID = -1
        matname = ""
        if hasattr(context, "material"):
            if hasattr(context.material, "name"):
                matname = context.material.name
                nummats = bpy.data.materials.__len__()
                for i in range(nummats):
                    if matname == bpy.data.materials[i].name:
                        matID = i
						
        if matID > -1:
            export_material(self.filepath, matID)

        return {"FINISHED"}

def menu_func_export(self, context):
    self.layout.operator(NOXMaterialExporter.bl_idname, text="NOX Material Export (*.nxm)")
    
#----------------------------------------------------------------

# This hack for writexml indentation problems was taken from
# http://ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-and-silly-whitespace/
    
def fixed_writexml(self, writer, indent="", addindent="", newl=""):
    # indent = current indentation
    # addindent = indentation to add to higher levels
    # newl = newline string
    writer.write(indent+"<" + self.tagName)

    attrs = self._get_attributes()
    a_names = attrs.keys()
    #a_names.sort()

    for a_name in a_names:
        writer.write(" %s=\"" % a_name)
        xml.dom.minidom._write_data(writer, attrs[a_name].value)
        writer.write("\"")
    if self.childNodes:
        if len(self.childNodes) == 1 \
          and self.childNodes[0].nodeType == xml.dom.minidom.Node.TEXT_NODE:
            writer.write(">")
            self.childNodes[0].writexml(writer, "", "", "")
            writer.write("</%s>%s" % (self.tagName, newl))
            return
        writer.write(">%s"%(newl))
        for node in self.childNodes:
            node.writexml(writer,indent+addindent,addindent,newl)
        writer.write("%s</%s>%s" % (indent,self.tagName,newl))
    else:
        writer.write("/>%s"%(newl))
        
# replace minidom's function with ours
    
#----------------------------------------------------------------

